from __future__ import annotations
import asyncio
import copy
import logging
from threading import Event
from typing import Dict, Optional, List, Tuple, TYPE_CHECKING, Sequence, Iterator, cast
from contextlib import contextmanager

from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
    mutable_configurations,
    pipette_definition,
)

from opentrons import types
from opentrons.config.types import RobotConfig
from opentrons.config import get_opentrons_path
from opentrons.drivers.smoothie_drivers import SimulatingDriver

from opentrons.drivers.rpi_drivers.gpio_simulator import SimulatingGPIOCharDev
from opentrons.util.async_helpers import ensure_yield

from .. import modules
from ..types import BoardRevision, Axis
from ..module_control import AttachedModulesControl
from ..util import ot2_axis_to_string

if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import PipetteName, PipetteModel
    from ..dev_types import (
        AttachedPipette,
        AttachedInstruments,
        PipetteSpec,
        InstrumentHardwareConfigs,
    )
    from opentrons.drivers.rpi_drivers.dev_types import GPIODriverLike


MODULE_LOG = logging.getLogger(__name__)


class Simulator:
    """This is a subclass of hardware_control that only simulates the
    hardware actions. It is suitable for use on a dev machine or on
    a robot with no smoothie connected.
    """

    @classmethod
    async def build(
        cls,
        attached_instruments: Dict[types.Mount, Dict[str, Optional[str]]],
        attached_modules: List[str],
        config: RobotConfig,
        loop: asyncio.AbstractEventLoop,
        strict_attached_instruments: bool = True,
    ) -> Simulator:
        """Build the simulator.

        Use this factory method rather than the initializer to handle proper GPIO
        initialization.

        :param attached_instruments: A dictionary describing the instruments
                                     the simulator should consider attached.
                                     If this argument is specified and
                                     :py:meth:`get_attached_instruments` is
                                     called with expectations that do not
                                     match, the call fails. This is useful for
                                     making the simulator match the real
                                     hardware, for instance to check if a
                                     protocol asks for the right instruments.
                                     This dict should map mounts to either
                                     empty dicts or to dicts containing
                                     'model' and 'id' keys.
        :param attached_modules: A list of module model names (e.g.
                                 `'tempdeck'` or `'magdeck'`) representing
                                 modules the simulator should assume are
                                 attached. Like `attached_instruments`, used
                                 to make the simulator match the setup of the
                                 real hardware.
        :param config: The robot config to use
        :param loop: The asyncio event loop to use.
        :param strict_attached_instruments: This param changes the behavior of
                                            the instrument cache. If ``True``,
                                            (default), ``cache_instrument``
                                            calls requesting instruments not
                                            in ``attached_instruments`` will
                                            fail as if the instrument was not
                                            present. If ``False``, those calls
                                            will still pass but give a response
                                            version of 1, while calls
                                            requesting instruments that _are_
                                            present get the full number.
        """
        gpio = SimulatingGPIOCharDev("gpiochip0")
        await gpio.setup()
        return cls(
            attached_instruments,
            attached_modules,
            config,
            loop,
            gpio,
            strict_attached_instruments,
        )

    def __init__(
        self,
        attached_instruments: Dict[types.Mount, Dict[str, Optional[str]]],
        attached_modules: List[str],
        config: RobotConfig,
        loop: asyncio.AbstractEventLoop,
        gpio_chardev: GPIODriverLike,
        strict_attached_instruments: bool = True,
    ) -> None:
        """Initialize the simulator.

        Always prefer using :py:meth:`.build` to create an instance of this class. For
        more information on arguments, see that method. If you want to use this
        directly, you must pass in an initialized _and set up_ GPIO driver instance.
        """
        self.config = config
        self._loop = loop
        self._smoothie_driver = SimulatingDriver()
        self._gpio_chardev = gpio_chardev

        def _sanitize_attached_instrument(
            passed_ai: Optional[PipetteSpec] = None,
        ) -> PipetteSpec:
            if not passed_ai or not passed_ai.get("model"):
                return {"model": None, "id": None}
            if pipette_load_name.supported_pipette(
                cast("PipetteName", passed_ai["model"])
            ):
                if pipette_load_name.is_model(passed_ai["model"]):
                    return passed_ai
                else:
                    get_pip_model = pipette_load_name.convert_pipette_name(
                        cast("PipetteName", passed_ai["model"])
                    )
                    return {
                        "model": PipetteModel(str(get_pip_model)),
                        "id": passed_ai.get("id"),
                    }
            raise KeyError(
                "If you specify attached_instruments, the model "
                "should be pipette names or pipette models, but "
                f'{passed_ai["model"]} is not'
            )

        self._attached_instruments = {
            m: _sanitize_attached_instrument(attached_instruments.get(m))  # type: ignore
            for m in types.Mount.ot2_mounts()
        }
        self._stubbed_attached_modules = attached_modules
        self._position = copy.copy(self._smoothie_driver.homed_position)
        # Engaged axes start all true in smoothie for some reason so we
        # imitate that here
        # TODO(LC2642019) Create a simulating driver for smoothie instead of
        # using a flag

        self._engaged_axes = {ax: True for ax in self._smoothie_driver.homed_position}
        self._lights = {"button": False, "rails": False}
        self._run_flag = Event()
        self._run_flag.set()
        self._log = MODULE_LOG.getChild(repr(self))
        self._strict_attached = bool(strict_attached_instruments)
        self._board_revision = BoardRevision.OG
        # TODO (lc 05-12-2021) In a follow-up refactor that pulls the execution
        # manager responsbility into the controller/backend itself as opposed
        # to the hardware api controller.
        self._module_controls: Optional[AttachedModulesControl] = None

    async def get_serial_number(self) -> Optional[str]:
        return "simulator"

    @property
    def gpio_chardev(self) -> GPIODriverLike:
        return self._gpio_chardev

    @property
    def module_controls(self) -> AttachedModulesControl:
        if not self._module_controls:
            raise AttributeError("Module controls not found.")
        return self._module_controls

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl) -> None:
        self._module_controls = module_controls

    @ensure_yield
    async def update_position(self) -> Dict[str, float]:
        return self._position

    def _unhomed_axes(self, axes: Sequence[str]) -> List[str]:
        return list(
            axis
            for axis in axes
            if not self._smoothie_driver.homed_flags.get(axis, False)
        )

    def is_homed(self, axes: Sequence[str]) -> bool:
        return not any(self._unhomed_axes(axes))

    @ensure_yield
    async def move(
        self,
        target_position: Dict[str, float],
        home_flagged_axes: bool = True,
        speed: Optional[float] = None,
        axis_max_speeds: Optional[Dict[str, float]] = None,
    ) -> None:
        self._position.update(target_position)
        self._engaged_axes.update({ax: True for ax in target_position})

    @ensure_yield
    async def home(self, axes: Optional[List[str]] = None) -> Dict[str, float]:
        # driver_3_0-> HOMED_POSITION
        checked_axes = "".join(axes) if axes else "XYZABC"
        self._position.update(
            {ax: self._smoothie_driver.homed_position[ax] for ax in checked_axes}
        )
        self._engaged_axes.update({ax: True for ax in checked_axes})
        await self._smoothie_driver.home(axis=checked_axes)
        return self._position

    @ensure_yield
    async def fast_home(self, axis: Sequence[str], margin: float) -> Dict[str, float]:
        for ax in axis:
            self._position[ax] = self._smoothie_driver.homed_position[ax]
            self._engaged_axes[ax] = True
        return self._position

    def _attached_to_mount(
        self, mount: types.Mount, expected_instr: Optional["PipetteName"]
    ) -> AttachedPipette:
        init_instr = self._attached_instruments.get(mount, {"model": None, "id": None})
        found_model = init_instr["model"]
        back_compat: List["PipetteName"] = []

        path_to_overrides = get_opentrons_path("pipette_config_overrides_dir")
        found_model_configs: Optional[pipette_definition.PipetteConfigurations] = None

        converted_found_name: Optional[pipette_definition.PipetteNameType] = None
        if found_model:
            converted_found_model = pipette_load_name.convert_pipette_model(found_model)
            converted_found_name = pipette_load_name.convert_to_pipette_name_type(
                found_model
            )
            found_model_configs = (
                mutable_configurations.load_with_mutable_configurations(
                    converted_found_model, path_to_overrides, init_instr["id"]
                )
            )
            back_compat = found_model_configs.pipette_backcompat_names
        if (
            expected_instr
            and found_model
            and found_model_configs
            and (
                str(converted_found_name) != expected_instr
                and expected_instr not in back_compat
            )
        ):
            if self._strict_attached:
                raise RuntimeError(
                    "mount {}: expected instrument {} but got {}".format(
                        mount.name, expected_instr, found_model
                    )
                )
            else:
                converted_expected_model = pipette_load_name.convert_pipette_name(
                    expected_instr
                )
                return {
                    "config": mutable_configurations.load_with_mutable_configurations(
                        converted_expected_model, path_to_overrides
                    ),
                    "id": None,
                }
        elif found_model and found_model_configs:
            # Instrument detected matches instrument expected (note:
            # "instrument detected" means passed as an argument to the
            # constructor of this class) OR,
            # Instrument detected and no expected instrument specified
            return {
                "config": found_model_configs,
                "id": init_instr["id"],
            }
        elif expected_instr:
            # Expected instrument specified and no instrument detected
            converted_expected_model = pipette_load_name.convert_pipette_name(
                expected_instr
            )
            return {
                "config": mutable_configurations.load_with_mutable_configurations(
                    converted_expected_model, path_to_overrides
                ),
                "id": None,
            }
        else:
            # No instrument detected or expected
            return {"config": None, "id": None}

    @ensure_yield
    async def get_attached_instruments(
        self, expected: Dict[types.Mount, "PipetteName"]
    ) -> AttachedInstruments:
        """Update the internal cache of attached instruments.

        This method allows after-init-time specification of attached simulated
        instruments. The method will return
        - the instruments specified at init-time, or if those do not exists,
        - the instruments specified in expected, or if that is not passed,
        - nothing

        :param expected: A mapping of mount to instrument model prefixes. When
                         loading instruments from a prefix, we return the
                         lexically-first model that matches the prefix. If the
                         models specified in expected do not match the models
                         specified in the `attached_instruments` argument of
                         :py:meth:`__init__`, :py:attr:`RuntimeError` is
                         raised.
        :raises RuntimeError: If an instrument is expected but not found.
        :returns: A dict of mount to either instrument model names or `None`.
        """
        return {
            mount: self._attached_to_mount(mount, expected.get(mount))
            for mount in types.Mount.ot2_mounts()
        }

    def set_active_current(self, axis_currents: Dict[Axis, float]) -> None:
        pass

    @ensure_yield
    async def watch(self) -> None:
        new_mods_at_ports = [
            modules.ModuleAtPort(port=f"/dev/ot_module_sim_{mod}{str(idx)}", name=mod)
            for idx, mod in enumerate(self._stubbed_attached_modules)
        ]
        await self.module_controls.register_modules(new_mods_at_ports=new_mods_at_ports)

    @contextmanager
    def save_current(self) -> Iterator[None]:
        yield

    @property
    def axis_bounds(self) -> Dict[Axis, Tuple[float, float]]:
        """The (minimum, maximum) bounds for each axis."""
        return {
            Axis[ax]: (0, pos)
            for ax, pos in self._smoothie_driver.axis_bounds.items()
            if ax not in "BC"
        }

    @property
    def fw_version(self) -> Optional[str]:
        return "Virtual Smoothie"

    @ensure_yield
    async def update_fw_version(self) -> None:
        pass

    @property
    def board_revision(self) -> BoardRevision:
        return self._board_revision

    @ensure_yield
    async def update_firmware(
        self, filename: str, loop: asyncio.AbstractEventLoop, modeset: bool
    ) -> str:
        return "Did nothing (simulating)"

    def engaged_axes(self) -> Dict[str, bool]:
        return self._engaged_axes

    @ensure_yield
    async def disengage_axes(self, axes: List[str]) -> None:
        self._engaged_axes.update({ax: False for ax in axes})

    def set_lights(self, button: Optional[bool], rails: Optional[bool]) -> None:
        if button is not None:
            self._lights["button"] = button
        if rails is not None:
            self._lights["rails"] = rails

    def get_lights(self) -> Dict[str, bool]:
        return self._lights

    def pause(self) -> None:
        self._run_flag.clear()

    def resume(self) -> None:
        self._run_flag.set()

    @ensure_yield
    async def halt(self) -> None:
        self._run_flag.set()

    @ensure_yield
    async def hard_halt(self) -> None:
        self._run_flag.set()

    @ensure_yield
    async def probe(self, axis: str, distance: float) -> Dict[str, float]:
        self._position[axis.upper()] = self._position[axis.upper()] + distance
        return self._position

    @ensure_yield
    async def clean_up(self) -> None:
        pass

    @ensure_yield
    async def configure_mount(
        self, mount: types.Mount, config: InstrumentHardwareConfigs
    ) -> None:
        mount_axis = Axis.by_mount(mount)
        plunger_axis = Axis.of_plunger(mount)

        await self._smoothie_driver.update_steps_per_mm(
            {ot2_axis_to_string(plunger_axis): config["steps_per_mm"]}
        )
        await self._smoothie_driver.update_pipette_config(
            ot2_axis_to_string(mount_axis), {"home": config["home_pos"]}
        )
        await self._smoothie_driver.update_pipette_config(
            ot2_axis_to_string(plunger_axis), {"max_travel": config["max_travel"]}
        )
        self._smoothie_driver.set_dwelling_current(
            {ot2_axis_to_string(plunger_axis): config["idle_current"]}
        )
        ms = config["splits"]
        if ms:
            self._smoothie_driver.configure_splits_for(
                {ot2_axis_to_string(plunger_axis): ms}
            )



from __future__ import annotations

# this file defines types that require dev dependencies
# and are only relevant for static typechecking. this file should only
# be imported if typing.TYPE_CHECKING is True
from typing import Optional, Dict, List, Union

from typing_extensions import TypedDict, Literal

from opentrons.hardware_control.instruments.ot3.instrument_calibration import (
    GripperCalibrationOffset,
)
from opentrons_shared_data.pipette.dev_types import (
    PipetteModel,
    PipetteName,
    ChannelCount,
)
from opentrons_shared_data.pipette.types import PipetteTipType
from opentrons_shared_data.pipette.pipette_definition import (
    PipetteConfigurations,
    SupportedTipsDefinition,
    PipetteBoundingBoxOffsetDefinition,
)
from opentrons_shared_data.gripper import (
    GripperModel,
    GripperDefinition,
)

from opentrons.drivers.types import MoveSplit
from opentrons.types import Mount
from opentrons.hardware_control.types import GripperJawState
from opentrons.hardware_control.nozzle_manager import NozzleMap


class InstrumentSpec(TypedDict):
    id: Optional[str]


class PipetteSpec(InstrumentSpec):
    model: Union[PipetteModel, None]


class GripperSpec(InstrumentSpec):
    model: Union[GripperModel, None]


class AttachedPipette(TypedDict):
    config: Optional[PipetteConfigurations]
    id: Optional[str]


class AttachedGripper(TypedDict):
    config: Optional[GripperDefinition]
    id: Optional[str]


AttachedInstruments = Dict[Mount, AttachedPipette]

OT3AttachedInstruments = Union[AttachedPipette, AttachedGripper]

EIGHT_CHANNELS = Literal[8]
ONE_CHANNEL = Literal[1]


class InstrumentDict(TypedDict):
    display_name: str


class PipetteDict(InstrumentDict):
    name: PipetteName
    model: PipetteModel
    back_compat_names: List[PipetteName]
    pipette_id: str
    min_volume: float
    max_volume: float
    channels: ChannelCount
    aspirate_flow_rate: float
    dispense_flow_rate: float
    blow_out_flow_rate: float
    aspirate_speed: float
    dispense_speed: float
    blow_out_speed: float
    current_volume: float
    tip_length: float
    working_volume: float
    tip_overlap: Dict[str, float]
    available_volume: float
    return_tip_height: float
    default_aspirate_flow_rates: Dict[str, float]
    default_dispense_flow_rates: Dict[str, float]
    default_blow_out_flow_rates: Dict[str, float]
    default_aspirate_speeds: Dict[str, float]
    default_dispense_speeds: Dict[str, float]
    default_blow_out_speeds: Dict[str, float]
    ready_to_aspirate: bool
    has_tip: bool
    default_push_out_volume: Optional[float]
    supported_tips: Dict[PipetteTipType, SupportedTipsDefinition]
    pipette_bounding_box_offsets: PipetteBoundingBoxOffsetDefinition
    current_nozzle_map: NozzleMap


class PipetteStateDict(TypedDict):
    tip_detected: bool


class GripperDict(InstrumentDict):
    model: GripperModel
    gripper_id: str
    state: GripperJawState  # Can we call this jaw_state?
    calibration_offset: GripperCalibrationOffset


class InstrumentHardwareConfigs(TypedDict):
    steps_per_mm: float
    home_pos: float
    max_travel: float
    idle_current: float
    splits: Optional[MoveSplit]



from typing import Dict, Optional
from typing_extensions import Protocol

from opentrons_shared_data.pipette.dev_types import PipetteName
from opentrons.types import Mount
from .types import MountArgType

# TODO (lc 12-05-2022) This protocol has deviated from the OT3 api. We
# need to figure out how to combine them again in follow-up refactors.
from opentrons.hardware_control.instruments.ot2.pipette import Pipette
from ..dev_types import PipetteDict
from ..types import CriticalPoint


class InstrumentConfigurer(Protocol[MountArgType]):
    """A protocol specifying how to interact with instrument presence and detection."""

    def reset_instrument(self, mount: Optional[MountArgType] = None) -> None:
        """
        Reset the internal state of a pipette by its mount, without doing
        any lower level reconfiguration. This is useful to make sure that no
        settings changes from a protocol persist.

        mount: If specified, reset that mount. If not specified, reset both
        """
        ...

    async def cache_instruments(
        self,
        require: Optional[Dict[Mount, PipetteName]] = None,
    ) -> None:
        """
        Scan the attached instruments, take necessary configuration actions,
        and set up hardware controller internal state if necessary.

        require: If specified, the require should be a dict of mounts to
                 instrument names describing the instruments expected to be
                 present. This can save a subsequent call of attached_instruments
                 and also serves as the hook for the hardware simulator to decide
                 what is attached.
        raises RuntimeError: If an instrument is expected but not found.

        This function will only change the things that need to be changed.
        If the same pipette (by serial) or the same lack of pipette is
        observed on a mount before and after the scan, no action will be
        taken. That makes this function appropriate for setting up the
        robot for operation, but not for making sure that any previous
        settings changes have been reset. For the latter use case, use
        reset_instrument.
        """
        ...

    def get_attached_instruments(self) -> Dict[Mount, PipetteDict]:
        """Get the status dicts of the cached attached instruments.

        Also available as :py:meth:`get_attached_instruments`.

        This returns a dictified version of the
        hardware_control.instruments.pipette.Pipette as a dict keyed by
        the Mount to which the pipette is attached.
        If no pipette is attached on a given mount, the mount key will
        still be present but will have the value ``None``.

        Note that on the OT-2 this is only a query of a cached value;
        to actively scan for changes, use cache_instruments`. This process
        deactivates the OT-2's motors and should be used sparingly.
        """
        ...

    def get_attached_instrument(self, mount: MountArgType) -> PipetteDict:
        """Get the status dict of a single cached instrument.

        Return values and caveats are as get_attached_instruments.
        """
        ...

    @property
    def attached_instruments(self) -> Dict[Mount, PipetteDict]:
        return self.get_attached_instruments()

    def get_attached_pipettes(self) -> Dict[Mount, PipetteDict]:
        """Get the status dicts of cached attached pipettes.

        Works like get_attached_instruments but for pipettes only - on the Flex,
        there will be no gripper information here.
        """
        ...

    @property
    def attached_pipettes(self) -> Dict[Mount, PipetteDict]:
        return self.get_attached_pipettes()

    def calibrate_plunger(
        self,
        mount: MountArgType,
        top: Optional[float] = None,
        bottom: Optional[float] = None,
        blow_out: Optional[float] = None,
        drop_tip: Optional[float] = None,
    ) -> None:
        """
        Set calibration values for the pipette plunger.
        This can be called multiple times as the user sets each value,
        or you can set them all at once.
        :param top: Touching but not engaging the plunger.
        :param bottom: Must be above the pipette's physical hard-stop, while
        still leaving enough room for 'blow_out'
        :param blow_out: Plunger is pushed down enough to expel all liquids.
        :param drop_tip: Position that causes the tip to be released from the
        pipette
        """
        ...

    def set_flow_rate(
        self,
        mount: MountArgType,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        """Set a pipette's rate of liquid handling in flow rate units"""
        ...

    def set_pipette_speed(
        self,
        mount: MountArgType,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        """Set a pipette's rate of liquid handling in linear speed units."""
        ...

    def get_instrument_max_height(
        self,
        mount: MountArgType,
        critical_point: Optional[CriticalPoint] = None,
    ) -> float:
        """Return max achievable height of the attached instrument
        based on the current critical point
        """
        ...

    async def add_tip(self, mount: MountArgType, tip_length: float) -> None:
        """Inform the hardware that a tip is now attached to a pipette.

        This changes the critical point of the pipette to make sure that
        the end of the tip is what moves around, and allows liquid handling.
        """
        ...

    async def remove_tip(self, mount: MountArgType) -> None:
        """Inform the hardware that a tip is no longer attached to a pipette.

        This changes the critical point of the system to the end of the
        nozzle and prevents further liquid handling commands.
        """
        ...

    def set_current_tiprack_diameter(
        self, mount: MountArgType, tiprack_diameter: float
    ) -> None:
        """Inform the hardware of the diameter of the tiprack.

        This drives the magnitude of the shake commanded for pipettes that need
        a shake after dropping or picking up tips.
        """
        ...

    def set_working_volume(self, mount: MountArgType, tip_volume: float) -> None:
        """Inform the hardware how much volume a pipette can aspirate.

        This will set the limit of aspiration for the pipette, and is
        necessary for backcompatibility.
        """
        ...

    @property
    def hardware_instruments(self) -> Dict[Mount, Optional[Pipette]]:
        """Return the underlying hardware representation of the instruments.

        This should rarely be used. Do not write new code that uses it.
        """
        ...

    def has_gripper(self) -> bool:
        """Return whether there is a gripper attached to this instance.

         - On robots that do not support a gripper, this will always return False.
         - On robots that support a gripper, this will return based on the current
        presence of a gripper.
        """
        ...



"""Run control command side-effect logic."""
import asyncio

from ..state import StateStore
from ..actions import ActionDispatcher, PauseAction, PauseSource


class RunControlHandler:
    """Implementation logic for protocol run control."""

    _state_store: StateStore

    def __init__(
        self,
        state_store: StateStore,
        action_dispatcher: ActionDispatcher,
    ) -> None:
        """Initialize a RunControlHandler instance."""
        self._state_store = state_store
        self._action_dispatcher = action_dispatcher

    async def wait_for_resume(self) -> None:
        """Issue a PauseAction to the store, pausing the run."""
        if not self._state_store.config.ignore_pause:
            self._action_dispatcher.dispatch(PauseAction(source=PauseSource.PROTOCOL))
            await self._state_store.wait_for(
                condition=self._state_store.commands.get_is_running
            )

    async def wait_for_duration(self, seconds: float) -> None:
        """Delay protocol execution for a duration."""
        if not self._state_store.config.ignore_pause:
            await asyncio.sleep(seconds)



"""File identifier interface."""

import json
from dataclasses import dataclass
from typing import Any, Dict, Sequence, Union, Optional

import anyio

from opentrons_shared_data.robot.dev_types import RobotType
from opentrons_shared_data.errors.exceptions import EnumeratedError, PythonException

from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols import parse
from opentrons.protocols.types import MalformedPythonProtocolError, PythonProtocol

from .file_reader_writer import BufferedFile
from .protocol_files_invalid_error import ProtocolFilesInvalidError
from .protocol_source import Metadata


JsonDict = Dict[str, Any]


@dataclass(frozen=True)
class IdentifiedJsonMain:
    """A file identified as a JSON protocol's main .json file."""

    original_file: BufferedFile
    """The original file that this was identified from."""

    unvalidated_json: JsonDict
    """The parsed JSON contents.

    Believed, but not confirmed at this point, to conform to one of our JSON protocol
    schemas.
    """

    schema_version: int
    """The JSON protocol schema that this file is believed to conform to."""

    robot_type: RobotType
    """The type of robot on which this protocol is meant to run."""

    metadata: Metadata
    """The protocol metadata extracted from this file."""


@dataclass(frozen=True)
class IdentifiedPythonMain:
    """A file identified as a Python protocol's main .py file."""

    original_file: BufferedFile
    """The original file that this was identified from."""

    api_level: APIVersion
    """The Python Protocol API apiLevel declared by the Python source."""

    robot_type: RobotType
    """The type of robot on which this protocol is meant to run."""

    metadata: Metadata
    """The protocol metadata extracted from this file."""


@dataclass(frozen=True)
class IdentifiedLabwareDefinition:
    """A file identified as a labware definition."""

    original_file: BufferedFile
    """The original file that this was identified from."""

    unvalidated_json: JsonDict
    """The parsed JSON contents.

    Believed, but not confirmed at this point, to conform to our labware definition
    schema v2.
    """


@dataclass(frozen=True)
class IdentifiedData:
    """A file identified as a user-defined data file."""

    original_file: BufferedFile
    """The original file that this was identified from."""


IdentifiedFile = Union[
    IdentifiedJsonMain,
    IdentifiedPythonMain,
    IdentifiedLabwareDefinition,
    IdentifiedData,
]


class FileIdentificationError(ProtocolFilesInvalidError):
    """Raised when FileIdentifier detects an invalid file."""

    def __init__(
        self,
        message: str,
        detail: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
        only_message: bool = False,
    ) -> None:
        super().__init__(message=message)
        self._only_message = only_message

    def __str__(self) -> str:
        """Special stringifier to conform to expecations about python protocol errors."""
        if self._only_message:
            return self.message
        else:
            return super().__str__()


class FileIdentifier:
    """File identifier interface."""

    @staticmethod
    async def identify(
        files: Sequence[BufferedFile], python_parse_mode: parse.PythonParseMode
    ) -> Sequence[IdentifiedFile]:
        """Identify the type and extract basic information from each file.

        This is intended to take ≲1 second per protocol on an OT-2, so it can extract
        basic information about all stored protocols relatively quickly. Fully parsing
        and validating protocols can take 10-100x longer, so that's left to other units,
        for only when it's really needed.
        """
        return [await _identify(file, python_parse_mode) for file in files]


async def _identify(
    file: BufferedFile, python_parse_mode: parse.PythonParseMode
) -> IdentifiedFile:
    lower_file_name = file.name.lower()
    if lower_file_name.endswith(".json"):
        return await _analyze_json(json_file=file)
    elif lower_file_name.endswith(".py"):
        return _analyze_python_protocol(
            py_file=file, python_parse_mode=python_parse_mode
        )
    elif lower_file_name.endswith(".csv") or lower_file_name.endswith(".txt"):
        return IdentifiedData(original_file=file)
    else:
        raise FileIdentificationError(
            message=f"{file.name} has an unrecognized file extension.",
            detail={"type": "bad-file-extension", "file": file.name},
        )


async def _analyze_json(
    json_file: BufferedFile,
) -> Union[IdentifiedJsonMain, IdentifiedLabwareDefinition]:
    try:
        json_contents = await anyio.to_thread.run_sync(json.loads, json_file.contents)
    except json.JSONDecodeError as e:
        raise FileIdentificationError(
            message=f"{json_file.name} is not valid JSON. {str(e)}",
            detail={"type": "invalid-json", "file": json_file.name},
            wrapping=[PythonException(e)],
        ) from e

    if _json_seems_like_labware(json_contents):
        return IdentifiedLabwareDefinition(
            original_file=json_file,
            unvalidated_json=json_contents,
        )
    elif _json_seems_like_protocol(json_contents):
        return _analyze_json_protocol(
            original_file=json_file,
            json_contents=json_contents,
        )
    else:
        raise FileIdentificationError(
            message=f"{json_file.name} is not a known Opentrons format.",
            detail={"type": "no-schema-match", "file": json_file.name},
        )


def _json_seems_like_labware(json: JsonDict) -> bool:
    # "ordering" and "wells" are required properties in our labware schema v2.
    return "ordering" in json and "wells" in json


def _json_seems_like_protocol(json: JsonDict) -> bool:
    # "schemaVersion" and "commands" are required properties in all of our JSON
    # protocol schemas since schema v3. (v7 is the latest at the time of writing.)
    #
    # When we stop supporting v3 files, we can look at "$otSharedSchema" instead,
    # which is more precise.
    return "schemaVersion" in json and "commands" in json


def _analyze_json_protocol(
    original_file: BufferedFile, json_contents: JsonDict
) -> IdentifiedJsonMain:
    error_message = f"{original_file.name} is not a valid JSON protocol."

    try:
        metadata = json_contents["metadata"]
        schema_version = json_contents["schemaVersion"]
        robot_type = json_contents["robot"]["model"]
    except KeyError as e:
        raise FileIdentificationError(
            message=error_message,
            detail={"kind": "missing-json-metadata", "missing-key": str(e)},
            wrapping=[PythonException(e)],
        ) from e

    # todo(mm, 2022-12-22): A JSON protocol file's metadata is not quite just an
    # arbitrary dict: its fields are supposed to follow a schema. Should we validate
    # this metadata against that schema instead of doing this simple isinstance() check?
    if not isinstance(metadata, dict):
        raise FileIdentificationError(
            message=error_message, detail={"kind": "json-metadata-not-object"}
        )

    if not isinstance(schema_version, int):
        raise FileIdentificationError(
            message=error_message,
            detail={
                "kind": "json-schema-version-not-int",
                "schema-version": schema_version,
            },
        )

    if robot_type not in ("OT-2 Standard", "OT-3 Standard"):
        raise FileIdentificationError(
            message=error_message,
            detail={"kind": "bad-json-protocol-robot-type", "robot-type": robot_type},
        )

    return IdentifiedJsonMain(
        original_file=original_file,
        unvalidated_json=json_contents,
        schema_version=schema_version,
        robot_type=robot_type,
        metadata=metadata,
    )


def _analyze_python_protocol(
    py_file: BufferedFile,
    python_parse_mode: parse.PythonParseMode,
) -> IdentifiedPythonMain:
    try:
        parsed = parse.parse(
            protocol_file=py_file.contents,
            filename=py_file.name,
            python_parse_mode=python_parse_mode,
        )
    except MalformedPythonProtocolError as e:
        raise FileIdentificationError(
            message=e.short_message,
            detail={"kind": "malformed-python-protocol"},
            wrapping=[PythonException(e)],
            only_message=True,
        ) from e

    # We know this should never be a JsonProtocol. Help out the type-checker.
    assert isinstance(
        parsed, PythonProtocol
    ), "Parsing a Python file returned something other than a Python protocol."

    if parsed.api_level > MAX_SUPPORTED_VERSION:
        raise FileIdentificationError(
            message=(
                f"API version {parsed.api_level} is not supported by this "
                f"robot software. Please either reduce your requested API "
                f"version or update your robot."
            ),
            detail={"kind": "future-api-version", "api-version": str(parsed.api_level)},
            only_message=True,
        )

    return IdentifiedPythonMain(
        original_file=py_file,
        metadata=parsed.metadata or {},
        robot_type=parsed.robot_type,
        api_level=parsed.api_level,
    )



"""Protocol API module implementation logic."""
from __future__ import annotations

from typing import Optional, List

from opentrons.hardware_control import SynchronousAdapter, modules as hw_modules
from opentrons.hardware_control.modules.types import (
    ModuleModel,
    TemperatureStatus,
    MagneticStatus,
    ThermocyclerStep,
    SpeedStatus,
    module_model_from_string,
)
from opentrons.drivers.types import (
    HeaterShakerLabwareLatchStatus,
    ThermocyclerLidStatus,
)
from opentrons.types import DeckSlotName
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient
from opentrons.protocol_engine.errors.exceptions import (
    LabwareNotLoadedOnModuleError,
    NoMagnetEngageHeightError,
)

from opentrons.protocols.api_support.types import APIVersion

from ... import validation
from ..module import (
    AbstractModuleCore,
    AbstractTemperatureModuleCore,
    AbstractMagneticModuleCore,
    AbstractThermocyclerCore,
    AbstractHeaterShakerCore,
    AbstractMagneticBlockCore,
)
from .exceptions import InvalidMagnetEngageHeightError


class ModuleCore(AbstractModuleCore):
    """Module core logic implementation for Python protocols.
    Args:
        module_id: ProtocolEngine ID of the loaded modules.
    """

    def __init__(
        self,
        module_id: str,
        engine_client: ProtocolEngineClient,
        api_version: APIVersion,
        sync_module_hardware: SynchronousAdapter[hw_modules.AbstractModule],
    ) -> None:
        self._module_id = module_id
        self._engine_client = engine_client
        self._api_version = api_version
        self._sync_module_hardware = sync_module_hardware

    @property
    def api_version(self) -> APIVersion:
        """Get the api version protocol module target."""
        return self._api_version

    @property
    def module_id(self) -> str:
        """The module's unique ProtocolEngine ID."""
        return self._module_id

    def get_model(self) -> ModuleModel:
        """Get the module's model identifier."""
        return module_model_from_string(
            self._engine_client.state.modules.get_connected_model(self.module_id)
        )

    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._engine_client.state.modules.get_serial_number(self.module_id)

    def get_deck_slot(self) -> DeckSlotName:
        """Get the module's deck slot."""
        return self._engine_client.state.modules.get_location(self.module_id).slotName

    def get_deck_slot_id(self) -> str:
        slot_name = self.get_deck_slot()
        return validation.internal_slot_to_public_string(
            slot_name, robot_type=self._engine_client.state.config.robot_type
        )

    def get_display_name(self) -> str:
        """Get the module's display name."""
        return self._engine_client.state.modules.get_definition(
            self.module_id
        ).displayName


class NonConnectedModuleCore(AbstractModuleCore):
    """Not connected module core logic implementation for Python protocols.

    Args:
        module_id: ProtocolEngine ID of the loaded modules.
    """

    def __init__(
        self,
        module_id: str,
        engine_client: ProtocolEngineClient,
        api_version: APIVersion,
    ) -> None:
        self._module_id = module_id
        self._engine_client = engine_client
        self._api_version = api_version

    @property
    def api_version(self) -> APIVersion:
        """Get the api version protocol module target."""
        return self._api_version

    @property
    def module_id(self) -> str:
        """The module's unique ProtocolEngine ID."""
        return self._module_id

    def get_model(self) -> ModuleModel:
        """Get the module's model identifier."""
        return module_model_from_string(
            self._engine_client.state.modules.get_connected_model(self.module_id)
        )

    def get_deck_slot(self) -> DeckSlotName:
        """Get the module's deck slot."""
        return self._engine_client.state.modules.get_location(self.module_id).slotName

    def get_display_name(self) -> str:
        """Get the module's display name."""
        return self._engine_client.state.modules.get_definition(
            self.module_id
        ).displayName

    def get_deck_slot_id(self) -> str:
        slot_name = self.get_deck_slot()
        return validation.internal_slot_to_public_string(
            slot_name, robot_type=self._engine_client.state.config.robot_type
        )


class TemperatureModuleCore(ModuleCore, AbstractTemperatureModuleCore):
    """Temperature Module core logic implementation for Python protocols."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.TempDeck]

    def set_target_temperature(self, celsius: float) -> None:
        """Set the Temperature Module's target temperature in °C."""
        self._engine_client.temperature_module_set_target_temperature(
            module_id=self.module_id, celsius=celsius
        )

    def wait_for_target_temperature(self, celsius: Optional[float] = None) -> None:
        """Wait until the module's target temperature is reached.
        Specifying a value for ``celsius`` that is different than
        the module's current target temperature may behave unpredictably.
        """
        self._engine_client.temperature_module_wait_for_target_temperature(
            module_id=self.module_id, celsius=celsius
        )

    def deactivate(self) -> None:
        """Deactivate the Temperature Module."""
        self._engine_client.temperature_module_deactivate(module_id=self.module_id)

    def get_current_temperature(self) -> float:
        """Get the module's current temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_target_temperature(self) -> Optional[float]:
        """Get the module's target temperature in °C, if set."""
        return self._sync_module_hardware.target  # type: ignore[no-any-return]

    def get_status(self) -> TemperatureStatus:
        """Get the module's current temperature status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]


class MagneticModuleCore(ModuleCore, AbstractMagneticModuleCore):
    """Magnetic Module control interface via a ProtocolEngine."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.MagDeck]

    def engage(
        self,
        height_from_base: Optional[float] = None,
        height_from_home: Optional[float] = None,
    ) -> None:
        """Raise the module's magnets.
        Only one of `height_from_base` or `height_from_home` may be specified.
        Args:
            height_from_base: Distance from labware base to raise the magnets.
            height_from_home: Distance from motor home position to raise the magnets.
        """

        # This core will only be used in apiLevels >=2.14, where
        # MagneticModuleContext.engage(height=...) is no longer available.
        # So these asserts should always pass.
        assert (
            height_from_home is None
        ), "Expected engage height to be specified from base."
        assert (
            height_from_base is not None
        ), "Expected engage height to be specified from base."

        self._engine_client.magnetic_module_engage(
            module_id=self._module_id, engage_height=height_from_base
        )

    def engage_to_labware(
        self, offset: float = 0, preserve_half_mm: bool = False
    ) -> None:
        """Raise the module's magnets up to its loaded labware.
        Args:
            offset: Offset from the labware's default engage height.
            preserve_half_mm: For labware whose definitions
                erroneously use half-mm for their defined default engage height,
                use the value directly instead of converting it to real millimeters.
        """
        try:
            default_height = (
                self._engine_client.state.labware.get_default_magnet_height(
                    module_id=self.module_id, offset=offset
                )
            )
        except LabwareNotLoadedOnModuleError:
            raise InvalidMagnetEngageHeightError(
                "There is no labware loaded on this Magnetic Module,"
                " so you must specify an engage height"
                " with the `height_from_base` parameter."
            )
        except NoMagnetEngageHeightError:
            raise InvalidMagnetEngageHeightError(
                "The labware loaded on this Magnetic Module"
                " does not have a default engage height,"
                " so you must specify an engage height"
                " with the `height_from_base` parameter."
            )

        self._engine_client.magnetic_module_engage(
            module_id=self.module_id, engage_height=default_height
        )

    def disengage(self) -> None:
        """Lower the magnets back into the module."""
        self._engine_client.magnetic_module_disengage(module_id=self.module_id)

    def get_status(self) -> MagneticStatus:
        """Get the module's current magnet status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]


class ThermocyclerModuleCore(ModuleCore, AbstractThermocyclerCore):
    """Core control interface for an attached Thermocycler Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.Thermocycler]
    _repetitions: Optional[int] = None
    _step_count: Optional[int] = None

    def open_lid(self) -> ThermocyclerLidStatus:
        """Open the Thermocycler's lid."""
        self._engine_client.thermocycler_open_lid(module_id=self.module_id)
        return ThermocyclerLidStatus.OPEN

    def close_lid(self) -> ThermocyclerLidStatus:
        """Close the Thermocycler's lid."""
        self._engine_client.thermocycler_close_lid(module_id=self.module_id)
        return ThermocyclerLidStatus.CLOSED

    def set_target_block_temperature(
        self,
        celsius: float,
        hold_time_seconds: Optional[float] = None,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Set the target temperature for the well block, in °C."""
        self._engine_client.thermocycler_set_target_block_temperature(
            module_id=self.module_id,
            celsius=celsius,
            block_max_volume=block_max_volume,
            hold_time_seconds=hold_time_seconds,
        )

    def wait_for_block_temperature(self) -> None:
        """Wait for target block temperature to be reached."""
        self._engine_client.thermocycler_wait_for_block_temperature(
            module_id=self.module_id
        )

    def set_target_lid_temperature(self, celsius: float) -> None:
        """Set the target temperature for the heated lid, in °C."""
        self._engine_client.thermocycler_set_target_lid_temperature(
            module_id=self.module_id, celsius=celsius
        )

    def wait_for_lid_temperature(self) -> None:
        """Wait for target lid temperature to be reached."""
        self._engine_client.thermocycler_wait_for_lid_temperature(
            module_id=self.module_id
        )

    def execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Execute a Thermocycler Profile."""
        self._repetitions = repetitions
        self._step_count = len(steps)
        self._engine_client.thermocycler_run_profile(
            module_id=self.module_id,
            steps=steps * repetitions,
            block_max_volume=block_max_volume,
        )

    def deactivate_lid(self) -> None:
        """Turn off the heated lid."""
        self._engine_client.thermocycler_deactivate_lid(module_id=self.module_id)

    def deactivate_block(self) -> None:
        """Turn off the well block temperature controller"""
        self._clear_cycle_counters()
        self._engine_client.thermocycler_deactivate_block(module_id=self.module_id)

    def deactivate(self) -> None:
        """Turn off the well block temperature controller, and heated lid"""
        self.deactivate_block()
        self.deactivate_lid()

    def get_lid_position(self) -> Optional[ThermocyclerLidStatus]:
        """Get the thermocycler's lid position."""
        return self._sync_module_hardware.lid_status  # type: ignore[no-any-return]

    def get_block_temperature_status(self) -> TemperatureStatus:
        """Get the thermocycler's block temperature status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]

    def get_lid_temperature_status(self) -> Optional[TemperatureStatus]:
        """Get the thermocycler's lid temperature status."""
        return self._sync_module_hardware.lid_temp_status  # type: ignore[no-any-return]

    def get_block_temperature(self) -> Optional[float]:
        """Get the thermocycler's current block temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_block_target_temperature(self) -> Optional[float]:
        """Get the thermocycler's target block temperature in °C."""
        return self._sync_module_hardware.target  # type: ignore[no-any-return]

    def get_lid_temperature(self) -> Optional[float]:
        """Get the thermocycler's current lid temperature in °C."""
        return self._sync_module_hardware.lid_temp  # type: ignore[no-any-return]

    def get_lid_target_temperature(self) -> Optional[float]:
        """Get the thermocycler's target lid temperature in °C."""
        return self._sync_module_hardware.lid_target  # type: ignore[no-any-return]

    def get_ramp_rate(self) -> Optional[float]:
        """Get the thermocycler's current ramp rate in °C/sec."""
        return self._sync_module_hardware.ramp_rate  # type: ignore[no-any-return]

    def get_hold_time(self) -> Optional[float]:
        """Get the remaining hold time in seconds."""
        return self._sync_module_hardware.hold_time  # type: ignore[no-any-return]

    def get_total_cycle_count(self) -> Optional[int]:
        """Get number of repetitions for current set cycle."""
        return self._repetitions

    def get_current_cycle_index(self) -> Optional[int]:
        """Get index of the current set cycle repetition."""
        if self._repetitions is None:
            return None
        step_index = self._sync_module_hardware.current_step_index
        # TODO(jbl 2022-10-31) this is intended to work even if execute profile is non-blocking, but it is blocking so
        #   this is not guaranteed to be accurate
        return (step_index - 1) // self._step_count + 1  # type: ignore[no-any-return]

    def get_total_step_count(self) -> Optional[int]:
        """Get number of steps within the current cycle."""
        return self._step_count

    def get_current_step_index(self) -> Optional[int]:
        """Get the index of the current step within the current cycle."""
        if self._step_count is None:
            return None
        step_index = self._sync_module_hardware.current_step_index
        # TODO(jbl 2022-10-31) this is intended to work even if execute profile is non-blocking, but it is blocking so
        #   this is not guaranteed to be accurate
        return (step_index - 1) % self._step_count + 1  # type: ignore[no-any-return]

    def _clear_cycle_counters(self) -> None:
        """Clear core-tracked cycle counters."""
        self._repetitions = None
        self._step_count = None


class HeaterShakerModuleCore(ModuleCore, AbstractHeaterShakerCore):
    """Core control interface for an attached Heater-Shaker Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.HeaterShaker]

    def set_target_temperature(self, celsius: float) -> None:
        """Set the labware plate's target temperature in °C."""
        self._engine_client.heater_shaker_set_target_temperature(
            module_id=self.module_id, celsius=celsius
        )

    def wait_for_target_temperature(self) -> None:
        """Wait for the labware plate's target temperature to be reached."""
        self._engine_client.heater_shaker_wait_for_temperature(module_id=self.module_id)

    def set_and_wait_for_shake_speed(self, rpm: int) -> None:
        """Set the shaker's target shake speed and wait for it to spin up."""
        self._engine_client.heater_shaker_set_and_wait_for_shake_speed(
            module_id=self.module_id, rpm=rpm
        )

    def open_labware_latch(self) -> None:
        """Open the labware latch."""
        self._engine_client.heater_shaker_open_labware_latch(module_id=self.module_id)

    def close_labware_latch(self) -> None:
        """Close the labware latch."""
        self._engine_client.heater_shaker_close_labware_latch(module_id=self.module_id)

    def deactivate_shaker(self) -> None:
        """Stop shaking."""
        self._engine_client.heater_shaker_deactivate_shaker(module_id=self.module_id)

    def deactivate_heater(self) -> None:
        """Stop heating."""
        self._engine_client.heater_shaker_deactivate_heater(module_id=self.module_id)

    def get_current_temperature(self) -> float:
        """Get the labware plate's current temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_target_temperature(self) -> Optional[float]:
        """Get the labware plate's target temperature in °C, if set."""
        return self._sync_module_hardware.target_temperature  # type: ignore[no-any-return]

    def get_current_speed(self) -> int:
        """Get the shaker's current speed in RPM."""
        return self._sync_module_hardware.speed  # type: ignore[no-any-return]

    def get_target_speed(self) -> Optional[int]:
        """Get the shaker's target speed in RPM, if set."""
        return self._sync_module_hardware.target_speed  # type: ignore[no-any-return]

    def get_temperature_status(self) -> TemperatureStatus:
        """Get the module's heater status."""
        return self._sync_module_hardware.temperature_status  # type: ignore[no-any-return]

    def get_speed_status(self) -> SpeedStatus:
        """Get the module's heater status."""
        return self._sync_module_hardware.speed_status  # type: ignore[no-any-return]

    def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        """Get the module's labware latch status."""
        return self._sync_module_hardware.labware_latch_status  # type: ignore[no-any-return]


class MagneticBlockCore(NonConnectedModuleCore, AbstractMagneticBlockCore):
    """Magnetic Block control interface via a ProtocolEngine."""



"""Command models to start heating a Thermocycler's block."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


SetTargetBlockTemperatureCommandType = Literal["thermocycler/setTargetBlockTemperature"]


class SetTargetBlockTemperatureParams(BaseModel):
    """Input parameters to set a Thermocycler's target block temperature."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler Module.")
    celsius: float = Field(..., description="Target temperature in °C.")
    blockMaxVolumeUl: Optional[float] = Field(
        None,
        description="Amount of liquid in uL of the most-full well"
        " in labware loaded onto the thermocycler.",
    )
    holdTimeSeconds: Optional[float] = Field(
        None,
        description="Amount of time, in seconds, to hold the temperature for."
        " If specified, a waitForBlockTemperature command will block until"
        " the given hold time has elapsed.",
    )


class SetTargetBlockTemperatureResult(BaseModel):
    """Result data from setting a Thermocycler's target block temperature."""

    targetBlockTemperature: float = Field(
        ...,
        description="The target block temperature that was set after validation.",
    )


class SetTargetBlockTemperatureImpl(
    AbstractCommandImpl[
        SetTargetBlockTemperatureParams,
        SetTargetBlockTemperatureResult,
    ]
):
    """Execution implementation of a Thermocycler's set block temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self,
        params: SetTargetBlockTemperatureParams,
    ) -> SetTargetBlockTemperatureResult:
        """Set a Thermocycler's target block temperature."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        target_temperature = thermocycler_state.validate_target_block_temperature(
            params.celsius
        )
        target_volume: Optional[float]
        if params.blockMaxVolumeUl is not None:
            target_volume = thermocycler_state.validate_max_block_volume(
                params.blockMaxVolumeUl
            )
        else:
            target_volume = None
        hold_time: Optional[float]
        if params.holdTimeSeconds is not None:
            hold_time = thermocycler_state.validate_hold_time(params.holdTimeSeconds)
        else:
            hold_time = None

        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        if thermocycler_hardware is not None:
            await thermocycler_hardware.set_target_block_temperature(
                target_temperature, volume=target_volume, hold_time_seconds=hold_time
            )

        return SetTargetBlockTemperatureResult(
            targetBlockTemperature=target_temperature
        )


class SetTargetBlockTemperature(
    BaseCommand[SetTargetBlockTemperatureParams, SetTargetBlockTemperatureResult]
):
    """A command to set a Thermocycler's target block temperature."""

    commandType: SetTargetBlockTemperatureCommandType = (
        "thermocycler/setTargetBlockTemperature"
    )
    params: SetTargetBlockTemperatureParams
    result: Optional[SetTargetBlockTemperatureResult]

    _ImplementationCls: Type[
        SetTargetBlockTemperatureImpl
    ] = SetTargetBlockTemperatureImpl


class SetTargetBlockTemperatureCreate(
    BaseCommandCreate[SetTargetBlockTemperatureParams]
):
    """A request to create a Thermocycler's set block temperature command."""

    commandType: SetTargetBlockTemperatureCommandType = (
        "thermocycler/setTargetBlockTemperature"
    )
    params: SetTargetBlockTemperatureParams

    _CommandCls: Type[SetTargetBlockTemperature] = SetTargetBlockTemperature



"""Models and implementation for the ``moveLabware`` command."""

from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons.types import Point
from ..types import (
    LabwareLocation,
    DeckSlotLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    LabwareMovementStrategy,
    LabwareOffsetVector,
    LabwareMovementOffsetData,
)
from ..errors import LabwareMovementNotAllowedError, NotSupportedOnRobotType
from ..resources import labware_validation, fixture_validation
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate
from opentrons_shared_data.gripper.constants import GRIPPER_PADDLE_WIDTH

if TYPE_CHECKING:
    from ..execution import EquipmentHandler, RunControlHandler, LabwareMovementHandler
    from ..state import StateView


MoveLabwareCommandType = Literal["moveLabware"]


# Extra buffer on top of minimum distance to move to the right
_TRASH_CHUTE_DROP_BUFFER_MM = 8


# TODO (spp, 2022-12-14): https://opentrons.atlassian.net/browse/RLAB-237
class MoveLabwareParams(BaseModel):
    """Input parameters for a ``moveLabware`` command."""

    labwareId: str = Field(..., description="The ID of the labware to move.")
    newLocation: LabwareLocation = Field(..., description="Where to move the labware.")
    strategy: LabwareMovementStrategy = Field(
        ...,
        description="Whether to use the gripper to perform the labware movement"
        " or to perform a manual movement with an option to pause.",
    )
    pickUpOffset: Optional[LabwareOffsetVector] = Field(
        None,
        description="Offset to use when picking up labware. "
        "Experimental param, subject to change",
    )
    dropOffset: Optional[LabwareOffsetVector] = Field(
        None,
        description="Offset to use when dropping off labware. "
        "Experimental param, subject to change",
    )


class MoveLabwareResult(BaseModel):
    """The output of a successful ``moveLabware`` command."""

    offsetId: Optional[str] = Field(
        # Default `None` instead of `...` so this field shows up as non-required in
        # OpenAPI. The server is allowed to omit it or make it null.
        None,
        description=(
            "An ID referencing the labware offset that will apply to this labware"
            " now that it's in the new location."
            " This offset will be in effect until the labware is moved"
            " with another `moveLabware` command."
            " Null or undefined means no offset applies,"
            " so the default of (0, 0, 0) will be used."
        ),
    )


class MoveLabwareImplementation(
    AbstractCommandImpl[MoveLabwareParams, MoveLabwareResult]
):
    """The execution implementation for ``moveLabware`` commands."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        labware_movement: LabwareMovementHandler,
        run_control: RunControlHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._labware_movement = labware_movement
        self._run_control = run_control

    async def execute(  # noqa: C901
        self, params: MoveLabwareParams
    ) -> MoveLabwareResult:
        """Move a loaded labware to a new location."""
        # Allow propagation of LabwareNotLoadedError.
        current_labware = self._state_view.labware.get(labware_id=params.labwareId)
        current_labware_definition = self._state_view.labware.get_definition(
            labware_id=params.labwareId
        )
        definition_uri = current_labware.definitionUri
        post_drop_slide_offset: Optional[Point] = None

        if self._state_view.labware.is_fixed_trash(params.labwareId):
            raise LabwareMovementNotAllowedError(
                f"Cannot move fixed trash labware '{current_labware_definition.parameters.loadName}'."
            )

        if isinstance(params.newLocation, AddressableAreaLocation):
            area_name = params.newLocation.addressableAreaName
            if not fixture_validation.is_gripper_waste_chute(
                area_name
            ) and not fixture_validation.is_deck_slot(area_name):
                raise LabwareMovementNotAllowedError(
                    f"Cannot move {current_labware.loadName} to addressable area {area_name}"
                )
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                area_name
            )

            if fixture_validation.is_gripper_waste_chute(area_name):
                # When dropping off labware in the waste chute, some bigger pieces
                # of labware (namely tipracks) can get stuck between a gripper
                # paddle and the bottom of the waste chute, even after the gripper
                # has homed all the way to the top of its travel. We add a "post-drop
                # slide" to dropoffs in the waste chute in order to guarantee that the
                # labware can drop fully through the chute before the gripper jaws close.
                post_drop_slide_offset = Point(
                    x=(current_labware_definition.dimensions.xDimension / 2.0)
                    + (GRIPPER_PADDLE_WIDTH / 2.0)
                    + _TRASH_CHUTE_DROP_BUFFER_MM,
                    y=0,
                    z=0,
                )
        elif isinstance(params.newLocation, DeckSlotLocation):
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                params.newLocation.slotName.id
            )

        available_new_location = self._state_view.geometry.ensure_location_not_occupied(
            location=params.newLocation
        )

        # Check that labware and destination do not have labware on top
        self._state_view.labware.raise_if_labware_has_labware_on_top(
            labware_id=params.labwareId
        )
        if isinstance(available_new_location, OnLabwareLocation):
            self._state_view.labware.raise_if_labware_has_labware_on_top(
                available_new_location.labwareId
            )
            # Ensure that labware can be placed on requested labware
            self._state_view.labware.raise_if_labware_cannot_be_stacked(
                top_labware_definition=current_labware_definition,
                bottom_labware_id=available_new_location.labwareId,
            )

        # Allow propagation of ModuleNotLoadedError.
        new_offset_id = self._equipment.find_applicable_labware_offset_id(
            labware_definition_uri=definition_uri,
            labware_location=available_new_location,
        )
        await self._labware_movement.ensure_movement_not_obstructed_by_module(
            labware_id=params.labwareId, new_location=available_new_location
        )

        if params.strategy == LabwareMovementStrategy.USING_GRIPPER:
            if self._state_view.config.robot_type == "OT-2 Standard":
                raise NotSupportedOnRobotType(
                    message="Labware movement using a gripper is not supported on the OT-2",
                    details={"strategy": params.strategy},
                )
            if not labware_validation.validate_gripper_compatible(
                current_labware_definition
            ):
                raise LabwareMovementNotAllowedError(
                    f"Cannot move labware '{current_labware_definition.parameters.loadName}' with gripper."
                    f" If trying to move a labware on an adapter, load the adapter separately to allow"
                    f" gripper movement."
                )
            if labware_validation.validate_definition_is_adapter(
                current_labware_definition
            ):
                raise LabwareMovementNotAllowedError(
                    f"Cannot move adapter '{current_labware_definition.parameters.loadName}' with gripper."
                )

            validated_current_loc = (
                self._state_view.geometry.ensure_valid_gripper_location(
                    current_labware.location
                )
            )
            validated_new_loc = self._state_view.geometry.ensure_valid_gripper_location(
                available_new_location,
            )
            user_offset_data = LabwareMovementOffsetData(
                pickUpOffset=params.pickUpOffset or LabwareOffsetVector(x=0, y=0, z=0),
                dropOffset=params.dropOffset or LabwareOffsetVector(x=0, y=0, z=0),
            )

            # Skips gripper moves when using virtual gripper
            await self._labware_movement.move_labware_with_gripper(
                labware_id=params.labwareId,
                current_location=validated_current_loc,
                new_location=validated_new_loc,
                user_offset_data=user_offset_data,
                post_drop_slide_offset=post_drop_slide_offset,
            )
        elif params.strategy == LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE:
            # Pause to allow for manual labware movement
            await self._run_control.wait_for_resume()

        return MoveLabwareResult(offsetId=new_offset_id)


class MoveLabware(BaseCommand[MoveLabwareParams, MoveLabwareResult]):
    """A ``moveLabware`` command."""

    commandType: MoveLabwareCommandType = "moveLabware"
    params: MoveLabwareParams
    result: Optional[MoveLabwareResult]

    _ImplementationCls: Type[MoveLabwareImplementation] = MoveLabwareImplementation


class MoveLabwareCreate(BaseCommandCreate[MoveLabwareParams]):
    """A request to create a ``moveLabware`` command."""

    commandType: MoveLabwareCommandType = "moveLabware"
    params: MoveLabwareParams

    _CommandCls: Type[MoveLabware] = MoveLabware



""" Type definitions for modules in this tree """
from dataclasses import dataclass
from typing import Dict, NamedTuple, Optional
from enum import Enum


class MoveSplit(NamedTuple):
    split_distance: float
    split_current: float
    split_speed: float
    after_time: float
    fullstep: bool


MoveSplits = Dict[str, MoveSplit]
#: Dict mapping axes to their split parameters


@dataclass
class Temperature:
    """Tempdeck, thermocycler plate, and heater-shaker temperatures"""

    current: float
    target: Optional[float]


@dataclass
class RPM:
    """Heater-shaker plate RPM"""

    current: int
    target: Optional[int]


class HeaterShakerLabwareLatchStatus(str, Enum):
    """Heater-shaker labware latch status"""

    OPENING = "opening"
    IDLE_OPEN = "idle_open"
    CLOSING = "closing"
    IDLE_CLOSED = "idle_closed"
    IDLE_UNKNOWN = "idle_unknown"
    UNKNOWN = "unknown"


@dataclass
class PlateTemperature(Temperature):
    """Thermocycler lid temperature"""

    hold: Optional[float]


class ThermocyclerLidStatus(str, Enum):
    """Thermocycler lid status."""

    UNKNOWN = "unknown"
    CLOSED = "closed"
    IN_BETWEEN = "in_between"
    OPEN = "open"
    MAX = "max"



"""Models for concrete occurrences of specific errors."""
from logging import getLogger

from datetime import datetime
from typing import Any, Dict, List, Type, Union, Optional, Sequence
from pydantic import BaseModel, Field
from opentrons_shared_data.errors.codes import ErrorCodes
from .exceptions import ProtocolEngineError
from opentrons_shared_data.errors.exceptions import EnumeratedError

log = getLogger(__name__)


# TODO(mc, 2021-11-12): flesh this model out with structured error data
# for each error type so client may produce better error messages
class ErrorOccurrence(BaseModel):
    """An occurrence of a specific error during protocol execution."""

    @classmethod
    def from_failed(
        cls: Type["ErrorOccurrence"],
        id: str,
        createdAt: datetime,
        error: Union[ProtocolEngineError, EnumeratedError],
    ) -> "ErrorOccurrence":
        """Build an ErrorOccurrence from the details available from a FailedAction or FinishAction."""
        if isinstance(error, ProtocolCommandFailedError) and error.original_error:
            wrappedErrors = [error.original_error]
        else:
            wrappedErrors = [
                cls.from_failed(id, createdAt, err) for err in error.wrapping
            ]
        return cls.construct(
            id=id,
            createdAt=createdAt,
            errorType=type(error).__name__,
            detail=error.message or str(error),
            errorInfo=error.detail,
            errorCode=error.code.value.code,
            wrappedErrors=wrappedErrors,
        )

    id: str = Field(..., description="Unique identifier of this error occurrence.")
    createdAt: datetime = Field(..., description="When the error occurred.")

    errorCode: str = Field(
        default=ErrorCodes.GENERAL_ERROR.value.code,
        description=(
            "An enumerated error code for the error type."
            " This is intended to be shown to the robot operator to direct them to the"
            " correct rough area for troubleshooting."
        ),
    )

    # TODO(mm, 2023-09-07):
    # The Opentrons App and Flex ODD use `errorType` in the title of error modals, but it's unclear
    # if they should. Is this field redundant now that we have `errorCode` and `detail`?
    #
    # In practice, this is often the source code name of our Python exception type.
    # Should we derive this from `errorCode` instead? Like how HTTP code 404 is always "not found."
    errorType: str = Field(
        ...,
        description=(
            "A short name for the error type that occurred, like `PipetteOverpressure`."
            " This can be a bit more specific than `errorCode`."
        ),
    )

    detail: str = Field(
        ...,
        description=(
            "A short human-readable message about the error."
            "\n\n"
            "This is intended to provide the robot operator with more specific details than"
            " `errorCode` alone. It should be no longer than a couple of sentences,"
            " and it should not contain internal newlines or indentation."
            "\n\n"
            " It should not internally repeat `errorCode`, but it may internally repeat `errorType`"
            " if it helps the message make sense when it's displayed in its own separate block."
        ),
    )

    errorInfo: Dict[str, str] = Field(
        default={},
        description=(
            "Specific details about the error that may be useful for determining cause."
            " This might contain the same information as `detail`, but in a more structured form."
            " It might also contain additional information that was too verbose or technical"
            " to put in `detail`."
        ),
    )

    wrappedErrors: List["ErrorOccurrence"] = Field(
        default=[], description="Errors that may have caused this one."
    )

    class Config:
        """Customize configuration for this model."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: object) -> None:
            """Append the schema to make the errorCode appear required.

            `errorCode`, `wrappedErrors`, and `errorInfo` have defaults because they are not included in earlier
            versions of this model, _and_ this model is loaded directly from
            the on-robot store. That means that, without a default, it will
            fail to parse. Once a default is defined, the automated schema will
            mark this as a non-required field, which is misleading as this is
            a response from the server to the client and it will always have an
            errorCode defined. This hack is required because it informs the client
            that it does not, in fact, have to account for a missing errorCode, wrappedError, or errorInfo.
            """
            schema["required"].extend(["errorCode", "wrappedErrors", "errorInfo"])


# TODO (tz, 7-12-23): move this to exceptions.py when we stop relaying on ErrorOccurrence.
class ProtocolCommandFailedError(ProtocolEngineError):
    """Raised if a fatal command execution error has occurred."""

    def __init__(
        self,
        original_error: Optional[ErrorOccurrence] = None,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ProtocolCommandFailedError."""
        super().__init__(ErrorCodes.GENERAL_ERROR, message, details, wrapping)
        self.original_error = original_error


ErrorOccurrence.update_forward_refs()



"""Heater-Shaker Module protocol commands."""

from .wait_for_temperature import (
    WaitForTemperature,
    WaitForTemperatureCreate,
    WaitForTemperatureParams,
    WaitForTemperatureResult,
    WaitForTemperatureCommandType,
)

from .set_target_temperature import (
    SetTargetTemperature,
    SetTargetTemperatureCreate,
    SetTargetTemperatureParams,
    SetTargetTemperatureResult,
    SetTargetTemperatureCommandType,
)

from .deactivate_heater import (
    DeactivateHeater,
    DeactivateHeaterCreate,
    DeactivateHeaterParams,
    DeactivateHeaterResult,
    DeactivateHeaterCommandType,
)

from .set_and_wait_for_shake_speed import (
    SetAndWaitForShakeSpeed,
    SetAndWaitForShakeSpeedCreate,
    SetAndWaitForShakeSpeedParams,
    SetAndWaitForShakeSpeedResult,
    SetAndWaitForShakeSpeedCommandType,
)

from .deactivate_shaker import (
    DeactivateShaker,
    DeactivateShakerCreate,
    DeactivateShakerParams,
    DeactivateShakerResult,
    DeactivateShakerCommandType,
)

from .open_labware_latch import (
    OpenLabwareLatch,
    OpenLabwareLatchCreate,
    OpenLabwareLatchParams,
    OpenLabwareLatchResult,
    OpenLabwareLatchCommandType,
)

from .close_labware_latch import (
    CloseLabwareLatch,
    CloseLabwareLatchCreate,
    CloseLabwareLatchParams,
    CloseLabwareLatchResult,
    CloseLabwareLatchCommandType,
)

__all__ = [
    # heaterShaker/waitForTemperature
    "WaitForTemperature",
    "WaitForTemperatureCreate",
    "WaitForTemperatureParams",
    "WaitForTemperatureResult",
    "WaitForTemperatureCommandType",
    # heaterShaker/setTargetTemperature
    "SetTargetTemperature",
    "SetTargetTemperatureCreate",
    "SetTargetTemperatureParams",
    "SetTargetTemperatureResult",
    "SetTargetTemperatureCommandType",
    # heaterShaker/deactivateHeater
    "DeactivateHeater",
    "DeactivateHeaterCreate",
    "DeactivateHeaterParams",
    "DeactivateHeaterResult",
    "DeactivateHeaterCommandType",
    # heaterShaker/setAndWaitForShakeSpeed
    "SetAndWaitForShakeSpeed",
    "SetAndWaitForShakeSpeedCreate",
    "SetAndWaitForShakeSpeedParams",
    "SetAndWaitForShakeSpeedResult",
    "SetAndWaitForShakeSpeedCommandType",
    # heaterShaker/deactivateShaker
    "DeactivateShaker",
    "DeactivateShakerCreate",
    "DeactivateShakerParams",
    "DeactivateShakerResult",
    "DeactivateShakerCommandType",
    # heaterShaker/openLabwareLatch
    "OpenLabwareLatch",
    "OpenLabwareLatchCreate",
    "OpenLabwareLatchParams",
    "OpenLabwareLatchResult",
    "OpenLabwareLatchCommandType",
    # heaterShaker/closeLabwareLatch
    "CloseLabwareLatch",
    "CloseLabwareLatchCreate",
    "CloseLabwareLatchParams",
    "CloseLabwareLatchResult",
    "CloseLabwareLatchCommandType",
]



"""Command models to start heating a Thermocycler's lid."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


SetTargetLidTemperatureCommandType = Literal["thermocycler/setTargetLidTemperature"]


class SetTargetLidTemperatureParams(BaseModel):
    """Input parameters to set a Thermocycler's target lid temperature."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler Module.")
    celsius: float = Field(..., description="Target temperature in °C.")


class SetTargetLidTemperatureResult(BaseModel):
    """Result data from setting a Thermocycler's target lid temperature."""

    targetLidTemperature: float = Field(
        ...,
        description="The target lid temperature that was set after validation.",
    )


class SetTargetLidTemperatureImpl(
    AbstractCommandImpl[SetTargetLidTemperatureParams, SetTargetLidTemperatureResult]
):
    """Execution implementation of a Thermocycler's set lid temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self,
        params: SetTargetLidTemperatureParams,
    ) -> SetTargetLidTemperatureResult:
        """Set a Thermocycler's target lid temperature."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        target_temperature = thermocycler_state.validate_target_lid_temperature(
            params.celsius
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        if thermocycler_hardware is not None:
            await thermocycler_hardware.set_target_lid_temperature(target_temperature)

        return SetTargetLidTemperatureResult(targetLidTemperature=target_temperature)


class SetTargetLidTemperature(
    BaseCommand[SetTargetLidTemperatureParams, SetTargetLidTemperatureResult]
):
    """A command to set a Thermocycler's target lid temperature."""

    commandType: SetTargetLidTemperatureCommandType = (
        "thermocycler/setTargetLidTemperature"
    )
    params: SetTargetLidTemperatureParams
    result: Optional[SetTargetLidTemperatureResult]

    _ImplementationCls: Type[SetTargetLidTemperatureImpl] = SetTargetLidTemperatureImpl


class SetTargetLidTemperatureCreate(BaseCommandCreate[SetTargetLidTemperatureParams]):
    """A request to create a Thermocycler's set lid temperature command."""

    commandType: SetTargetLidTemperatureCommandType = (
        "thermocycler/setTargetLidTemperature"
    )
    params: SetTargetLidTemperatureParams

    _CommandCls: Type[SetTargetLidTemperature] = SetTargetLidTemperature



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from opentrons.hardware_control.modules import ModuleModel as HardwareModuleModel
from opentrons.protocol_engine import ProtocolEngine, LabwareOffsetLocation, ModuleModel
from opentrons.types import DeckSlotName, Point

from ..labware import LabwareLoadParams


@dataclass
class ProvidedLabwareOffset:
    """A labware offset provided externally.

    Parameters:
        delta: The positional adjustment that should apply to all movements
            to this labware. Measured in deck coordinates, from the nominal
            position to the adjusted position.
        offset_id: An ID referencing the relevant external offset resource.
            `None` means no matching offset.
    """

    delta: Point
    offset_id: Optional[str]


class AbstractLabwareOffsetProvider(ABC):
    @abstractmethod
    def find(
        self,
        load_params: LabwareLoadParams,
        requested_module_model: Optional[HardwareModuleModel],
        deck_slot: DeckSlotName,
    ) -> ProvidedLabwareOffset:
        """Return the offset that should apply to a newly loaded labware.

        An APIv2 protocol's `ProtocolContext` should call this once for each labware,
        as it loads it.

        Args:
            labware_definition_uri: The labware's definition URI.
            requested_module_model: If the labware is atop a module,
                the model of that module.
                To ensure stability between simulation and execution,
                this is the model that the protocol requested,
                not the model that was actually found via the hardware API.
                (They can be different because of module compatibility.)
            deck_slot: The deck slot that the labware occupies. Or, if the labware is
                atop a module, the deck slot that the module occupies.
        """
        ...


class NullLabwareOffsetProvider(AbstractLabwareOffsetProvider):
    """Always provides (0, 0, 0)."""

    def find(
        self,
        load_params: LabwareLoadParams,
        requested_module_model: Optional[HardwareModuleModel],
        deck_slot: DeckSlotName,
    ) -> ProvidedLabwareOffset:
        return ProvidedLabwareOffset(delta=Point(0, 0, 0), offset_id=None)


class LabwareOffsetProvider(AbstractLabwareOffsetProvider):
    """Provides a `ProtocolEngine`'s labware offsets."""

    def __init__(self, engine: ProtocolEngine) -> None:
        """Initialize an offset provider with access to ProtocolEngine state."""
        self._labware_view = engine.state_view.labware

    def find(
        self,
        load_params: LabwareLoadParams,
        requested_module_model: Optional[HardwareModuleModel],
        deck_slot: DeckSlotName,
    ) -> ProvidedLabwareOffset:
        """Look up an offset in ProtocolEngine state and return it, if one exists.

        See the parent class for param details.
        """
        offset = self._labware_view.find_applicable_labware_offset(
            definition_uri=load_params.as_uri(),
            location=LabwareOffsetLocation(
                slotName=deck_slot,
                moduleModel=(
                    None
                    if requested_module_model is None
                    else ModuleModel(requested_module_model.value)
                ),
            ),
        )
        if offset is None:
            return ProvidedLabwareOffset(
                delta=Point(x=0, y=0, z=0),
                offset_id=None,
            )
        else:
            return ProvidedLabwareOffset(
                delta=Point(x=offset.vector.x, y=offset.vector.y, z=offset.vector.z),
                offset_id=offset.id,
            )



"""A Protocol-Engine-friendly wrapper for opentrons.motion_planning.deck_conflict."""
from __future__ import annotations
import itertools
import logging
from typing import (
    Collection,
    Dict,
    Optional,
    Tuple,
    overload,
    Union,
    TYPE_CHECKING,
    List,
)

from opentrons_shared_data.errors.exceptions import MotionPlanningFailureError
from opentrons_shared_data.module import FLEX_TC_LID_COLLISION_ZONE

from opentrons.hardware_control.nozzle_manager import NozzleConfigurationType
from opentrons.hardware_control.modules.types import ModuleType
from opentrons.motion_planning import deck_conflict as wrapped_deck_conflict
from opentrons.motion_planning import adjacent_slots_getters

from opentrons.protocol_engine import (
    StateView,
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    OFF_DECK_LOCATION,
    WellLocation,
    DropTipWellLocation,
)
from opentrons.protocol_engine.errors.exceptions import LabwareNotLoadedOnModuleError
from opentrons.protocol_engine.types import (
    StagingSlotLocation,
)
from opentrons.types import DeckSlotName, StagingSlotName, Point
from . import point_calculations
from ..._trash_bin import TrashBin
from ..._waste_chute import WasteChute

if TYPE_CHECKING:
    from ...labware import Labware


class PartialTipMovementNotAllowedError(MotionPlanningFailureError):
    """Error raised when trying to perform a partial tip movement to an illegal location."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message=message,
        )


class UnsuitableTiprackForPipetteMotion(MotionPlanningFailureError):
    """Error raised when trying to perform a pipette movement to a tip rack, based on adapter status."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message=message,
        )


_log = logging.getLogger(__name__)

# TODO (spp, 2023-12-06): move this to a location like motion planning where we can
#  derive these values from geometry definitions
#  Also, verify y-axis extents values for the nozzle columns.
# Bounding box measurements
A12_column_front_left_bound = Point(x=-11.03, y=2)
A12_column_back_right_bound = Point(x=526.77, y=506.2)

_NOZZLE_PITCH = 9
A1_column_front_left_bound = Point(
    x=A12_column_front_left_bound.x - _NOZZLE_PITCH * 11, y=2
)
A1_column_back_right_bound = Point(
    x=A12_column_back_right_bound.x - _NOZZLE_PITCH * 11, y=506.2
)

_FLEX_TC_LID_BACK_LEFT_PT = Point(
    x=FLEX_TC_LID_COLLISION_ZONE["back_left"]["x"],
    y=FLEX_TC_LID_COLLISION_ZONE["back_left"]["y"],
    z=FLEX_TC_LID_COLLISION_ZONE["back_left"]["z"],
)

_FLEX_TC_LID_FRONT_RIGHT_PT = Point(
    x=FLEX_TC_LID_COLLISION_ZONE["front_right"]["x"],
    y=FLEX_TC_LID_COLLISION_ZONE["front_right"]["y"],
    z=FLEX_TC_LID_COLLISION_ZONE["front_right"]["z"],
)


@overload
def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    new_labware_id: str,
) -> None:
    pass


@overload
def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    new_module_id: str,
) -> None:
    pass


@overload
def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    new_trash_bin: TrashBin,
) -> None:
    pass


def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    # TODO(mm, 2023-02-23): This interface is impossible to use correctly. In order
    # to have new_labware_id or new_module_id, the caller needs to have already loaded
    # the new item into Protocol Engine--but then, it's too late to do deck conflict.
    # checking. Find a way to do deck conflict checking before the new item is loaded.
    new_labware_id: Optional[str] = None,
    new_module_id: Optional[str] = None,
    new_trash_bin: Optional[TrashBin] = None,
) -> None:
    """Check for conflicts between items on the deck.

    This is a Protocol-Engine-friendly wrapper around
    opentrons.motion_planning.deck_conflict.check().

    Params:
        engine_state: An interface to retrieve details about the deck items.
        existing_labware_ids: The Protocol Engine IDs of all labware already loaded.
        existing_module_ids: The Protocol Engine IDs of all modules already loaded.
        new_labware_id: The Protocol Engine ID of a new labware you've just added.
            Mutually exclusive with new_module_id.
        new_module_id: The Protocol EngineID of a new module you've just added.
            Mutually exclusive with new_labware_id.

    Raises:
        opentrons.motion_planning.deck_conflict.DeckConflictError:
            If the newly-added item conflicts with one of the existing items.
    """

    if new_labware_id is not None:
        new_location_and_item = _map_labware(engine_state, new_labware_id)
    if new_module_id is not None:
        new_location_and_item = _map_module(engine_state, new_module_id)
    if new_trash_bin is not None:
        new_location_and_item = _map_disposal_location(new_trash_bin)

    if new_location_and_item is None:
        # The new item should be excluded from deck conflict checking. Nothing to do.
        return

    new_location, new_item = new_location_and_item

    all_existing_labware = (
        _map_labware(engine_state, labware_id) for labware_id in existing_labware_ids
    )
    mapped_existing_labware = (m for m in all_existing_labware if m is not None)

    all_existing_modules = (
        _map_module(engine_state, module_id) for module_id in existing_module_ids
    )
    mapped_existing_modules = (m for m in all_existing_modules if m is not None)

    all_exisiting_disposal_locations = (
        _map_disposal_location(disposal_location)
        for disposal_location in existing_disposal_locations
    )
    mapped_disposal_locations = (
        m for m in all_exisiting_disposal_locations if m is not None
    )

    existing_items: Dict[
        Union[DeckSlotName, StagingSlotName], wrapped_deck_conflict.DeckItem
    ] = {}
    for existing_location, existing_item in itertools.chain(
        mapped_existing_labware, mapped_existing_modules, mapped_disposal_locations
    ):
        assert existing_location not in existing_items
        existing_items[existing_location] = existing_item

    wrapped_deck_conflict.check(
        existing_items=existing_items,
        new_item=new_item,
        new_location=new_location,
        robot_type=engine_state.config.robot_type,
    )


# TODO (spp, 2023-02-16): move pipette movement safety checks to its own separate file.
def check_safe_for_pipette_movement(
    engine_state: StateView,
    pipette_id: str,
    labware_id: str,
    well_name: str,
    well_location: Union[WellLocation, DropTipWellLocation],
) -> None:
    """Check if the labware is safe to move to with a pipette in partial tip configuration.

    Args:
        engine_state: engine state view
        pipette_id: ID of the pipette to be moved
        labware_id: ID of the labware we are moving to
        well_name: Name of the well to move to
        well_location: exact location within the well to move to
    """
    # TODO (spp, 2023-02-06): remove this check after thorough testing.
    #  This function is capable of checking for movement conflict regardless of
    #  nozzle configuration.
    if not engine_state.pipettes.get_is_partially_configured(pipette_id):
        return

    if isinstance(well_location, DropTipWellLocation):
        # convert to WellLocation
        well_location = engine_state.geometry.get_checked_tip_drop_location(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_location=well_location,
            partially_configured=True,
        )
    well_location_point = engine_state.geometry.get_well_position(
        labware_id=labware_id, well_name=well_name, well_location=well_location
    )
    primary_nozzle = engine_state.pipettes.get_primary_nozzle(pipette_id)

    if not _is_within_pipette_extents(
        engine_state=engine_state, pipette_id=pipette_id, location=well_location_point
    ):
        raise PartialTipMovementNotAllowedError(
            f"Requested motion with the {primary_nozzle} nozzle partial configuration"
            f" is outside of robot bounds for the pipette."
        )

    labware_slot = engine_state.geometry.get_ancestor_slot_name(labware_id)
    pipette_bounds_at_well_location = (
        engine_state.pipettes.get_pipette_bounds_at_specified_move_to_position(
            pipette_id=pipette_id, destination_position=well_location_point
        )
    )
    surrounding_slots = adjacent_slots_getters.get_surrounding_slots(
        slot=labware_slot.as_int(), robot_type=engine_state.config.robot_type
    )

    if _will_collide_with_thermocycler_lid(
        engine_state=engine_state,
        pipette_bounds=pipette_bounds_at_well_location,
        surrounding_regular_slots=surrounding_slots.regular_slots,
    ):
        raise PartialTipMovementNotAllowedError(
            f"Moving to {engine_state.labware.get_display_name(labware_id)} in slot"
            f" {labware_slot} with {primary_nozzle} nozzle partial configuration"
            f" will result in collision with thermocycler lid in deck slot A1."
        )

    for regular_slot in surrounding_slots.regular_slots:
        if _slot_has_potential_colliding_object(
            engine_state=engine_state,
            pipette_bounds=pipette_bounds_at_well_location,
            surrounding_slot=regular_slot,
        ):
            raise PartialTipMovementNotAllowedError(
                f"Moving to {engine_state.labware.get_display_name(labware_id)} in slot"
                f" {labware_slot} with {primary_nozzle} nozzle partial configuration"
                f" will result in collision with items in deck slot {regular_slot}."
            )
    for staging_slot in surrounding_slots.staging_slots:
        if _slot_has_potential_colliding_object(
            engine_state=engine_state,
            pipette_bounds=pipette_bounds_at_well_location,
            surrounding_slot=staging_slot,
        ):
            raise PartialTipMovementNotAllowedError(
                f"Moving to {engine_state.labware.get_display_name(labware_id)} in slot"
                f" {labware_slot} with {primary_nozzle} nozzle partial configuration"
                f" will result in collision with items in staging slot {staging_slot}."
            )


def _slot_has_potential_colliding_object(
    engine_state: StateView,
    pipette_bounds: Tuple[Point, Point, Point, Point],
    surrounding_slot: Union[DeckSlotName, StagingSlotName],
) -> bool:
    """Return the slot, if any, that has an item that the pipette might collide into."""
    # Check if slot overlaps with pipette position
    slot_pos = engine_state.addressable_areas.get_addressable_area_position(
        addressable_area_name=surrounding_slot.id,
        do_compatibility_check=False,
    )
    slot_bounds = engine_state.addressable_areas.get_addressable_area_bounding_box(
        addressable_area_name=surrounding_slot.id,
        do_compatibility_check=False,
    )
    slot_back_left_coords = Point(slot_pos.x, slot_pos.y + slot_bounds.y, slot_pos.z)
    slot_front_right_coords = Point(slot_pos.x + slot_bounds.x, slot_pos.y, slot_pos.z)

    # If slot overlaps with pipette bounds
    if point_calculations.are_overlapping_rectangles(
        rectangle1=(pipette_bounds[0], pipette_bounds[1]),
        rectangle2=(slot_back_left_coords, slot_front_right_coords),
    ):
        # Check z-height of items in overlapping slot
        if isinstance(surrounding_slot, DeckSlotName):
            slot_highest_z = engine_state.geometry.get_highest_z_in_slot(
                DeckSlotLocation(slotName=surrounding_slot)
            )
        else:
            slot_highest_z = engine_state.geometry.get_highest_z_in_slot(
                StagingSlotLocation(slotName=surrounding_slot)
            )
        return slot_highest_z >= pipette_bounds[0].z
    return False


def _will_collide_with_thermocycler_lid(
    engine_state: StateView,
    pipette_bounds: Tuple[Point, Point, Point, Point],
    surrounding_regular_slots: List[DeckSlotName],
) -> bool:
    """Return whether the pipette might collide with thermocycler's lid/clips on a Flex.

    If any of the pipette's bounding vertices lie inside the no-go zone of the thermocycler-
    which is the area that's to the left, back and below the thermocycler's lid's
    protruding clips, then we will mark the movement for possible collision.

    This could cause false raises for the case where an 8-channel is accessing the
    thermocycler labware in a location such that the pipette is in the area between
    the clips but not touching either clips. But that's a tradeoff we'll need to make
    between a complicated check involving accurate positions of all entities involved
    and a crude check that disallows all partial tip movements around the thermocycler.
    """
    # TODO (spp, 2024-02-27): Improvements:
    #  - make the check dynamic according to lid state:
    #     - if lid is open, check if pipette is in no-go zone
    #     - if lid is closed, use the closed lid height to check for conflict
    if (
        DeckSlotName.SLOT_A1 in surrounding_regular_slots
        and engine_state.modules.is_flex_deck_with_thermocycler()
    ):
        return (
            point_calculations.are_overlapping_rectangles(
                rectangle1=(_FLEX_TC_LID_BACK_LEFT_PT, _FLEX_TC_LID_FRONT_RIGHT_PT),
                rectangle2=(pipette_bounds[0], pipette_bounds[1]),
            )
            and pipette_bounds[0].z <= _FLEX_TC_LID_BACK_LEFT_PT.z
        )

    return False


def check_safe_for_tip_pickup_and_return(
    engine_state: StateView,
    pipette_id: str,
    labware_id: str,
) -> None:
    """Check if the presence or absence of a tiprack adapter might cause any pipette movement issues.

    A 96 channel pipette will pick up tips using cam action when it's configured
    to use ALL nozzles. For this, the tiprack needs to be on the Flex 96 channel tiprack adapter
    or similar or the tips will not be picked up.

    On the other hand, if the pipette is configured with partial nozzle configuration,
    it uses the usual pipette presses to pick the tips up, in which case, having the tiprack
    on the Flex 96 channel tiprack adapter (or similar) will cause the pipette to
    crash against the adapter posts.

    In order to check if the 96-channel can move and pickup/drop tips safely, this method
    checks for the height attribute of the tiprack adapter rather than checking for the
    specific official adapter since users might create custom labware &/or definitions
    compatible with the official adapter.
    """
    if not engine_state.pipettes.get_channels(pipette_id) == 96:
        # Adapters only matter to 96 ch.
        return

    is_partial_config = engine_state.pipettes.get_is_partially_configured(pipette_id)
    tiprack_name = engine_state.labware.get_display_name(labware_id)
    tiprack_parent = engine_state.labware.get_location(labware_id)
    if isinstance(tiprack_parent, OnLabwareLocation):  # tiprack is on an adapter
        is_96_ch_tiprack_adapter = engine_state.labware.get_has_quirk(
            labware_id=tiprack_parent.labwareId, quirk="tiprackAdapterFor96Channel"
        )
        tiprack_height = engine_state.labware.get_dimensions(labware_id).z
        adapter_height = engine_state.labware.get_dimensions(tiprack_parent.labwareId).z
        if is_partial_config and tiprack_height < adapter_height:
            raise PartialTipMovementNotAllowedError(
                f"{tiprack_name} cannot be on an adapter taller than the tip rack"
                f" when picking up fewer than 96 tips."
            )
        elif not is_partial_config and not is_96_ch_tiprack_adapter:
            raise UnsuitableTiprackForPipetteMotion(
                f"{tiprack_name} must be on an Opentrons Flex 96 Tip Rack Adapter"
                f" in order to pick up or return all 96 tips simultaneously."
            )

    elif (
        not is_partial_config
    ):  # tiprack is not on adapter and pipette is in full config
        raise UnsuitableTiprackForPipetteMotion(
            f"{tiprack_name} must be on an Opentrons Flex 96 Tip Rack Adapter"
            f" in order to pick up or return all 96 tips simultaneously."
        )


# TODO (spp, 2023-02-06): update the extents check to use all nozzle bounds instead of
#  just position of primary nozzle when checking if the pipette is out-of-bounds
def _is_within_pipette_extents(
    engine_state: StateView,
    pipette_id: str,
    location: Point,
) -> bool:
    """Whether a given point is within the extents of a configured pipette on the specified robot."""
    robot_type = engine_state.config.robot_type
    pipette_channels = engine_state.pipettes.get_channels(pipette_id)
    nozzle_config = engine_state.pipettes.get_nozzle_layout_type(pipette_id)
    primary_nozzle = engine_state.pipettes.get_primary_nozzle(pipette_id)
    if robot_type == "OT-3 Standard":
        if pipette_channels == 96 and nozzle_config == NozzleConfigurationType.COLUMN:
            # TODO (spp, 2023-12-18): change this eventually to use column mappings in
            #  the pipette geometry definitions.
            if primary_nozzle == "A12":
                return (
                    A12_column_front_left_bound.x
                    <= location.x
                    <= A12_column_back_right_bound.x
                    and A12_column_front_left_bound.y
                    <= location.y
                    <= A12_column_back_right_bound.y
                )
            elif primary_nozzle == "A1":
                return (
                    A1_column_front_left_bound.x
                    <= location.x
                    <= A1_column_back_right_bound.x
                    and A1_column_front_left_bound.y
                    <= location.y
                    <= A1_column_back_right_bound.y
                )
    # TODO (spp, 2023-11-07): check for 8-channel nozzle A1 & H1 extents on Flex & OT2
    return True


def _map_labware(
    engine_state: StateView,
    labware_id: str,
) -> Optional[
    Tuple[Union[DeckSlotName, StagingSlotName], wrapped_deck_conflict.DeckItem]
]:
    location_from_engine = engine_state.labware.get_location(labware_id=labware_id)

    if isinstance(location_from_engine, AddressableAreaLocation):
        # This will be guaranteed to be either deck slot name or staging slot name
        slot: Union[DeckSlotName, StagingSlotName]
        try:
            slot = DeckSlotName.from_primitive(location_from_engine.addressableAreaName)
        except ValueError:
            slot = StagingSlotName.from_primitive(
                location_from_engine.addressableAreaName
            )
        return (
            slot,
            wrapped_deck_conflict.Labware(
                name_for_errors=engine_state.labware.get_load_name(
                    labware_id=labware_id
                ),
                highest_z=engine_state.geometry.get_labware_highest_z(
                    labware_id=labware_id
                ),
                uri=engine_state.labware.get_definition_uri(labware_id=labware_id),
                is_fixed_trash=engine_state.labware.is_fixed_trash(
                    labware_id=labware_id
                ),
            ),
        )

    elif isinstance(location_from_engine, DeckSlotLocation):
        # This labware is loaded directly into a deck slot.
        # Map it to a wrapped_deck_conflict.Labware.
        return (
            location_from_engine.slotName,
            wrapped_deck_conflict.Labware(
                name_for_errors=engine_state.labware.get_load_name(
                    labware_id=labware_id
                ),
                highest_z=engine_state.geometry.get_labware_highest_z(
                    labware_id=labware_id
                ),
                uri=engine_state.labware.get_definition_uri(labware_id=labware_id),
                is_fixed_trash=engine_state.labware.is_fixed_trash(
                    labware_id=labware_id
                ),
            ),
        )

    elif isinstance(location_from_engine, ModuleLocation):
        # This labware is loaded atop a module. Don't map it to anything here;
        # let _map_module() pick it up.
        return None

    elif isinstance(location_from_engine, OnLabwareLocation):
        # TODO(jbl 2023-06-08) check if we need to do any logic here or if this is correct
        return None

    elif location_from_engine == OFF_DECK_LOCATION:
        # This labware is off-deck. Exclude it from conflict checking.
        # todo(mm, 2023-02-23): Move this logic into wrapped_deck_conflict.
        return None


def _map_module(
    engine_state: StateView,
    module_id: str,
) -> Optional[Tuple[DeckSlotName, wrapped_deck_conflict.DeckItem]]:
    module_model = engine_state.modules.get_connected_model(module_id=module_id)
    module_type = module_model.as_type()
    mapped_location = engine_state.modules.get_location(module_id=module_id).slotName

    # Use the module model (e.g. "temperatureModuleV1") as the name for error messages
    # because it's convenient for us. Unfortunately, this won't necessarily match
    # the string that the Python protocol author used to load the module.
    name_for_errors = module_model.value

    highest_z_including_labware = _get_module_highest_z_including_labware(
        engine_state=engine_state,
        module_id=module_id,
    )

    if module_type == ModuleType.HEATER_SHAKER:
        return (
            mapped_location,
            wrapped_deck_conflict.HeaterShakerModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
            ),
        )
    elif module_type == ModuleType.MAGNETIC_BLOCK:
        return (
            mapped_location,
            wrapped_deck_conflict.MagneticBlockModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
            ),
        )
    elif module_type == ModuleType.THERMOCYCLER:
        return (
            mapped_location,
            wrapped_deck_conflict.ThermocyclerModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
                # Python Protocol API >=v2.14 never allows loading a Thermocycler in
                # its semi configuration.
                is_semi_configuration=False,
            ),
        )
    else:
        return (
            mapped_location,
            wrapped_deck_conflict.OtherModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
            ),
        )


def _map_disposal_location(
    disposal_location: Union[Labware, WasteChute, TrashBin],
) -> Optional[Tuple[DeckSlotName, wrapped_deck_conflict.DeckItem]]:
    if isinstance(disposal_location, TrashBin):
        return (
            disposal_location.location,
            wrapped_deck_conflict.TrashBin(name_for_errors="trash bin"),
        )
    else:
        return None


def _deck_slot_to_int(deck_slot_location: DeckSlotLocation) -> int:
    return deck_slot_location.slotName.as_int()


def _get_module_highest_z_including_labware(
    engine_state: StateView, module_id: str
) -> float:
    try:
        labware_id = engine_state.labware.get_id_by_module(module_id=module_id)
    except LabwareNotLoadedOnModuleError:
        # No labware is loaded atop this module.
        # The height should be just the module itself.
        return engine_state.modules.get_overall_height(module_id=module_id)
    else:
        # This module has a labware loaded atop it. The height should include both.
        return engine_state.geometry.get_labware_highest_z(labware_id=labware_id)



"""Utililty methods and classes for interacting with the Module Status Server."""

import asyncio
from typing import Sequence, Set, Callable, List, Awaitable

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control.emulation.module_server.client import (
    ModuleStatusClient,
    ModuleServerClientError,
    ModuleServerDisconnected,
)
from opentrons.hardware_control.emulation.module_server.models import Message
from opentrons.hardware_control.emulation.module_server.server import log
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.emulation.types import ModuleType
from opentrons.hardware_control.modules import ModuleAtPort

NotifyMethod = Callable[[List[ModuleAtPort], List[ModuleAtPort]], Awaitable[None]]
"""Signature of method to be notified of new and removed modules."""


async def listen_module_connection(callback: NotifyMethod) -> None:
    """Listen for module emulator connections."""
    settings = Settings()
    try:
        client = await ModuleStatusClient.connect(
            host=settings.module_server.host,
            port=settings.module_server.port,
            interval_seconds=1.0,
        )
        listener = ModuleListener(client=client, notify_method=callback)
        await listener.run()
    except IOError:
        log.exception("Failed to connect to module server.")


class ModuleListener:
    """Provide a callback for listening for new and removed module connections."""

    def __init__(self, client: ModuleStatusClient, notify_method: NotifyMethod) -> None:
        """Constructor.

        Args:
            client: A module server client
            notify_method: callback method.

        Returns:
            None
        """
        self._client = client
        self._notify_method = notify_method
        self._hub_index = 1

    async def run(self) -> None:
        """Run the listener."""
        while True:
            try:
                m = await self._client.read()
                await self.handle_message(message=m)
            except ModuleServerDisconnected:
                log.info("Disconnected from module server.")
                break
            except ModuleServerClientError:
                log.exception("Read error.")
                break

    async def handle_message(self, message: Message) -> None:
        """Call callback with results of message.

        Args:
            message: Message object from module server

        Returns:
            None
        """

        def _next_index() -> int:
            index = self._hub_index
            self._hub_index += 1
            return index

        connections = [
            ModuleAtPort(
                port=c.url,
                name=c.module_type,
                usb_port=USBPort(name=c.identifier, port_number=_next_index()),
            )
            for c in message.connections
        ]
        if message.status == "connected" or message.status == "dump":
            await self._notify_method(connections, [])
        elif message.status == "disconnected":
            await self._notify_method([], connections)


async def wait_emulators(
    client: ModuleStatusClient,
    modules: Sequence[ModuleType],
    timeout: float,
) -> None:
    """Wait for module emulators to connect.

    Args:
        client: module server client.
        modules: collection of of module types to wait for.
        timeout: how long to wait for emulators to connect (in seconds)

    Returns:
        None
    Raises:
        asyncio.TimeoutError on timeout.
    """

    async def _wait_modules() -> None:
        """Read messages from module server waiting for modules in module_set to
        be connected."""
        module_set: Set[str] = set(modules)

        while module_set:
            m: Message = await client.read()
            if m.status == "dump" or m.status == "connected":
                for c in m.connections:
                    module_set.discard(c.module_type)
            elif m.status == "disconnected":
                for c in m.connections:
                    module_set.add(c.module_type)

            log.debug(f"after message: {m}, awaiting module set is: {module_set}")

    await asyncio.wait_for(_wait_modules(), timeout=timeout)



"""
USB Simulating Driver.

A class to convert info from the usb bus into a
more readable format.
"""
from typing import List

from opentrons.hardware_control.modules.types import ModuleAtPort

from .interfaces import USBDriverInterface


class USBBusSimulator(USBDriverInterface):
    def match_virtual_ports(
        self, virtual_port: List[ModuleAtPort]
    ) -> List[ModuleAtPort]:
        return virtual_port



from __future__ import annotations

import asyncio
import logging
from typing import Optional, List

from opentrons.drivers.command_builder import CommandBuilder

from .errors import NoResponse, AlarmResponse, ErrorResponse
from .async_serial import AsyncSerial

log = logging.getLogger(__name__)


class SerialConnection:
    @classmethod
    async def _build_serial(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        loop: Optional[asyncio.AbstractEventLoop],
        reset_buffer_before_write: bool,
    ) -> AsyncSerial:
        return await AsyncSerial.create(
            port=port,
            baud_rate=baud_rate,
            timeout=timeout,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )

    @classmethod
    async def create(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        ack: str,
        name: Optional[str] = None,
        retry_wait_time_seconds: float = 0.1,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        error_keyword: Optional[str] = None,
        alarm_keyword: Optional[str] = None,
        reset_buffer_before_write: bool = False,
    ) -> SerialConnection:
        """
        Create a connection.

        Args:
            port: url or port to connect to
            baud_rate: baud rate
            timeout: timeout in seconds
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            loop: optional event loop.
            error_keyword: optional string that will cause an
                           ErrorResponse exception when detected
                           (default: error)
            alarm_keyword: optional string that will cause an
                           AlarmResponse exception when detected
                           (default: alarm)
            reset_buffer_before_write: whether to reset the read buffer before
              every write

        Returns: SerialConnection
        """
        serial = await cls._build_serial(
            port=port,
            baud_rate=baud_rate,
            timeout=timeout,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )
        name = name or port
        return cls(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
            error_keyword=error_keyword or "error",
            alarm_keyword=alarm_keyword or "alarm",
        )

    def __init__(
        self,
        serial: AsyncSerial,
        port: str,
        name: str,
        ack: str,
        retry_wait_time_seconds: float,
        error_keyword: str,
        alarm_keyword: str,
    ) -> None:
        """
        Constructor

        Args:
            serial: AsyncSerial object
            port: url or port to connect to
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            error_keyword: string that will cause an ErrorResponse
                           exception when detected
            alarm_keyword: string that will cause an AlarmResponse
                           exception when detected
        """
        self._serial = serial
        self._port = port
        self._name = name
        self._ack = ack.encode()
        self._retry_wait_time_seconds = retry_wait_time_seconds
        self._send_data_lock = asyncio.Lock()
        self._error_keyword = error_keyword.lower()
        self._alarm_keyword = alarm_keyword.lower()

    async def send_command(
        self, command: CommandBuilder, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send a command and return the response.

        Args:
            command: A command builder.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        return await self.send_data(
            data=command.build(), retries=retries, timeout=timeout
        )

    async def send_dfu_command(self, command: CommandBuilder) -> None:
        """
        Send a dfu command to enter device bootloader.

        This method doesn't wait for a response after sending the command since the
        module port gets disconnected once it enters its bootloader.
        """
        encoded_command = command.build().encode()

        async with self._send_data_lock:
            log.debug(f"{self.name}: Write -> {encoded_command!r}")
            await self._serial.write(data=encoded_command)

    async def send_data(
        self, data: str, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        async with self._send_data_lock, self._serial.timeout_override(
            "timeout", timeout
        ):
            return await self._send_data(data=data, retries=retries)

    async def _send_data(self, data: str, retries: int = 0) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout

        Returns: The command response

        Raises: SerialException
        """
        data_encode = data.encode()

        for retry in range(retries + 1):
            log.debug(f"{self.name}: Write -> {data_encode!r}")
            await self._serial.write(data=data_encode)

            response = await self._serial.read_until(match=self._ack)
            log.debug(f"{self.name}: Read <- {response!r}")

            if (
                self._ack in response
                or self._error_keyword.encode() in response.lower()
            ):
                # Remove ack from response
                response = response.replace(self._ack, b"")
                str_response = self.process_raw_response(
                    command=data, response=response.decode()
                )
                self.raise_on_error(response=str_response)
                return str_response

            log.info(f"{self.name}: retry number {retry}/{retries}")

            await self.on_retry()

        raise NoResponse(port=self._port, command=data)

    async def open(self) -> None:
        """Open the connection."""
        await self._serial.open()

    async def close(self) -> None:
        """Close the connection."""
        await self._serial.close()

    async def is_open(self) -> bool:
        """Check if connection is open."""
        return await self._serial.is_open()

    @property
    def port(self) -> str:
        return self._port

    @property
    def name(self) -> str:
        return self._name

    @property
    def send_data_lock(self) -> asyncio.Lock:
        return self._send_data_lock

    def raise_on_error(self, response: str) -> None:
        """
        Raise an error if the response contains an error

        Args:
            response: response

        Returns: None

        Raises: SerialException
        """
        lower = response.lower()

        if self._alarm_keyword in lower:
            raise AlarmResponse(port=self._port, response=response)

        if self._error_keyword in lower:
            raise ErrorResponse(port=self._port, response=response)

    async def on_retry(self) -> None:
        """
        Opportunity for derived classes to perform action between retries. Default
        behaviour is to wait then re-open the connection.

        Returns: None
        """
        await asyncio.sleep(self._retry_wait_time_seconds)
        await self._serial.close()
        await self._serial.open()

    def process_raw_response(self, command: str, response: str) -> str:
        """
        Opportunity for derived classes to process the raw response. Default
         strips white space.

        Args:
            command: The sent command.
            response: The raw read response minus ack.

        Returns:
            processed response.
        """
        return response.strip()


class AsyncResponseSerialConnection(SerialConnection):
    @classmethod
    async def create(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        ack: str,
        name: Optional[str] = None,
        retry_wait_time_seconds: float = 0.1,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        error_keyword: Optional[str] = None,
        alarm_keyword: Optional[str] = None,
        reset_buffer_before_write: bool = False,
        async_error_ack: Optional[str] = None,
    ) -> AsyncResponseSerialConnection:
        """
        Create a connection.

        Args:
            port: url or port to connect to
            baud_rate: baud rate
            timeout: timeout in seconds
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            loop: optional event loop.
            error_keyword: optional string that will cause an
                           ErrorResponse exception when detected
                           (default: error)
            alarm_keyword: optional string that will cause an
                           AlarmResponse exception when detected
                           (default: alarm)
            reset_buffer_before_write: whether to reset the read buffer before
              every write
            async_error_ack: optional string that will indicate an asynchronous
                             error when detected (default: async)

        Returns: AsyncResponseSerialConnection
        """
        serial = await super()._build_serial(
            port=port,
            baud_rate=baud_rate,
            timeout=timeout,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )
        name = name or port
        return cls(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
            error_keyword=error_keyword or "err",
            alarm_keyword=alarm_keyword or "alarm",
            async_error_ack=async_error_ack or "async",
        )

    def __init__(
        self,
        serial: AsyncSerial,
        port: str,
        name: str,
        ack: str,
        retry_wait_time_seconds: float,
        error_keyword: str,
        alarm_keyword: str,
        async_error_ack: str,
    ) -> None:
        """
        Constructor

        Args:
            serial: AsyncSerial object
            port: url or port to connect to
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            error_keyword: string that will cause an ErrorResponse
                           exception when detected
            alarm_keyword: string that will cause an AlarmResponse
                           exception when detected
            async_error_ack: string that will indicate an asynchronous
                             error when detected
        """
        super().__init__(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
            error_keyword=error_keyword,
            alarm_keyword=alarm_keyword,
        )
        self._serial = serial
        self._port = port
        self._name = name
        self._ack = ack.encode()
        self._retry_wait_time_seconds = retry_wait_time_seconds
        self._error_keyword = error_keyword.lower()
        self._alarm_keyword = alarm_keyword.lower()
        self._async_error_ack = async_error_ack.lower()

    async def send_command(
        self, command: CommandBuilder, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send a command and return the response.

        Args:
            command: A command builder.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        return await self.send_data(
            data=command.build(), retries=retries, timeout=timeout
        )

    async def send_data(
        self, data: str, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        async with super().send_data_lock, self._serial.timeout_override(
            "timeout", timeout
        ):
            return await self._send_data(data=data, retries=retries)

    async def _send_data(self, data: str, retries: int = 0) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout

        Returns: The command response

        Raises: SerialException
        """
        data_encode = data.encode()

        for retry in range(retries + 1):
            log.debug(f"{self._name}: Write -> {data_encode!r}")
            await self._serial.write(data=data_encode)

            response: List[bytes] = []
            response.append(await self._serial.read_until(match=self._ack))
            log.debug(f"{self._name}: Read <- {response[-1]!r}")

            while self._async_error_ack.encode() in response[-1].lower():
                # check for multiple a priori async errors
                response.append(await self._serial.read_until(match=self._ack))
                log.debug(f"{self._name}: Read <- {response[-1]!r}")

            for r in response:
                if self._async_error_ack.encode() in r:
                    # Remove ack from response
                    ackless_response = r.replace(self._ack, b"")
                    str_response = self.process_raw_response(
                        command=data, response=ackless_response.decode()
                    )
                    self.raise_on_error(response=str_response)

            if self._ack in response[-1]:
                # Remove ack from response
                ackless_response = response[-1].replace(self._ack, b"")
                str_response = self.process_raw_response(
                    command=data, response=ackless_response.decode()
                )
                self.raise_on_error(response=str_response)
                return str_response

            log.info(f"{self._name}: retry number {retry}/{retries}")

            await self.on_retry()

        raise NoResponse(port=self._port, command=data)



# noqa: D100

from __future__ import annotations
import logging
from typing import Callable, Dict, List
from typing_extensions import Literal

from opentrons.commands import types


MODULE_LOG = logging.getLogger(__name__)


class LegacyBroker:
    """A pub/sub message broker.

    Deprecated:
        Use the newer, more generic `opentrons.utils.Broker` class instead.
        This class is coupled to old types from `opentrons.commands`.
        https://opentrons.atlassian.net/browse/RSS-270
    """

    def __init__(self) -> None:
        self.subscriptions: Dict[
            Literal["command"],
            List[Callable[[types.CommandMessage], None]],
        ] = {}
        self.logger = MODULE_LOG

    def subscribe(  # noqa: D102
        self,
        topic: Literal["command"],
        handler: Callable[[types.CommandMessage], None],
    ) -> Callable[[], None]:
        def unsubscribe() -> None:
            try:
                self.subscriptions[topic].remove(handler)
            except ValueError:
                # handler already removed from list
                pass

        if handler not in self.subscriptions.setdefault(topic, []):
            self.subscriptions[topic].append(handler)

        return unsubscribe

    def publish(  # noqa: D102
        self, topic: Literal["command"], message: types.CommandMessage
    ) -> None:
        [handler(message) for handler in self.subscriptions.get(topic, [])]

    def set_logger(self, logger: logging.Logger) -> None:  # noqa: D102
        self.logger = logger



""" opentrons.calibration_storage.encoder_decoder: classes that
allow you to customize serialization to/from json.
"""
import json
import datetime
from typing import Any, cast


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class DateTimeDecoder(json.JSONDecoder):
    def __init__(self) -> None:
        super().__init__(object_hook=self.dict_to_obj)

    def dict_to_obj(self, d: object) -> Any:
        if isinstance(d, dict):
            d = {k: self._decode_datetime(v) for k, v in d.items()}
        return d

    def _decode_datetime(self, obj: object) -> Any:
        try:
            return datetime.datetime.fromisoformat(cast(str, obj))
        except ValueError:
            return obj
        except TypeError:
            return self.dict_to_obj(obj)



"""Move to well command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..types import DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    WellLocationMixin,
    MovementMixin,
    DestinationPositionResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import MovementHandler

MoveToWellCommandType = Literal["moveToWell"]


class MoveToWellParams(PipetteIdMixin, WellLocationMixin, MovementMixin):
    """Payload required to move a pipette to a specific well."""

    pass


class MoveToWellResult(DestinationPositionResult):
    """Result data from the execution of a MoveToWell command."""

    pass


class MoveToWellImplementation(AbstractCommandImpl[MoveToWellParams, MoveToWellResult]):
    """Move to well command implementation."""

    def __init__(self, movement: MovementHandler, **kwargs: object) -> None:
        self._movement = movement

    async def execute(self, params: MoveToWellParams) -> MoveToWellResult:
        """Move the requested pipette to the requested well."""
        x, y, z = await self._movement.move_to_well(
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            well_location=params.wellLocation,
            force_direct=params.forceDirect,
            minimum_z_height=params.minimumZHeight,
            speed=params.speed,
        )

        return MoveToWellResult(position=DeckPoint(x=x, y=y, z=z))


class MoveToWell(BaseCommand[MoveToWellParams, MoveToWellResult]):
    """Move to well command model."""

    commandType: MoveToWellCommandType = "moveToWell"
    params: MoveToWellParams
    result: Optional[MoveToWellResult]

    _ImplementationCls: Type[MoveToWellImplementation] = MoveToWellImplementation


class MoveToWellCreate(BaseCommandCreate[MoveToWellParams]):
    """Move to well command creation request model."""

    commandType: MoveToWellCommandType = "moveToWell"
    params: MoveToWellParams

    _CommandCls: Type[MoveToWell] = MoveToWell



"""Move relative (jog) command payload, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..types import MovementAxis, DeckPoint
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate
from .pipetting_common import DestinationPositionResult

if TYPE_CHECKING:
    from ..execution import MovementHandler


MoveRelativeCommandType = Literal["moveRelative"]


class MoveRelativeParams(BaseModel):
    """Payload required for a MoveRelative command."""

    pipetteId: str = Field(..., description="Pipette to move.")
    axis: MovementAxis = Field(..., description="Axis along which to move.")
    distance: float = Field(
        ...,
        description=(
            "Distance to move in millimeters. A positive number will move"
            " towards the right (x), back (y), top (z) of the deck."
        ),
    )


class MoveRelativeResult(DestinationPositionResult):
    """Result data from the execution of a MoveRelative command."""

    pass


class MoveRelativeImplementation(
    AbstractCommandImpl[MoveRelativeParams, MoveRelativeResult]
):
    """Move relative command implementation."""

    def __init__(self, movement: MovementHandler, **kwargs: object) -> None:
        self._movement = movement

    async def execute(self, params: MoveRelativeParams) -> MoveRelativeResult:
        """Move (jog) a given pipette a relative distance."""
        x, y, z = await self._movement.move_relative(
            pipette_id=params.pipetteId,
            axis=params.axis,
            distance=params.distance,
        )

        return MoveRelativeResult(position=DeckPoint(x=x, y=y, z=z))


class MoveRelative(BaseCommand[MoveRelativeParams, MoveRelativeResult]):
    """Command to move (jog) a given pipette a relative distance."""

    commandType: MoveRelativeCommandType = "moveRelative"
    params: MoveRelativeParams
    result: Optional[MoveRelativeResult]

    _ImplementationCls: Type[MoveRelativeImplementation] = MoveRelativeImplementation


class MoveRelativeCreate(BaseCommandCreate[MoveRelativeParams]):
    """Data to create a MoveRelative command."""

    commandType: MoveRelativeCommandType = "moveRelative"
    params: MoveRelativeParams

    _CommandCls: Type[MoveRelative] = MoveRelative
